#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: supervisor metrics reporting manager
#
import collections
import time

from snmpsim_control_plane import error
from snmpsim_control_plane import log
from snmpsim_control_plane.supervisor import lifecycle
from snmpsim_control_plane.supervisor.reporting import collector
from snmpsim_control_plane.supervisor.reporting.formats import jsondoc
from snmpsim_control_plane.supervisor.reporting.formats import null


class ReportingManager(object):
    """Gather and dump activity metrics.

    For each of given managed process instances, collect OS-level
    metrics.

    Then write them down as a JSON file indexed by time. Consumers
    are expected to process each of these files and are free to remove
    them.
    """

    REPORTING_PERIOD = 15

    REPORTERS = {
        'null': null.NullReporter,
        'jsondoc': jsondoc.JsonDocReporter,
    }

    STARTED = int(time.time())

    _last_reportings = collections.defaultdict(dict)

    _reporter = null.NullReporter()

    _next_dump = time.time() + REPORTING_PERIOD

    @classmethod
    def configure(cls, fmt, *args):
        try:
            reporter = cls.REPORTERS[fmt]

        except KeyError:
            raise error.ControlPlaneError(
                'Unsupported reporting format: %s' % fmt)

        cls._reporter = reporter(*args)

        log.info('Using "%s" activity reporting method with '
                 'params %s' % (cls._reporter, ', '.join(args)))

    @classmethod
    def process_metrics(cls, watch_dir, *instances):
        now = int(time.time())

        if cls._next_dump > now:
            return

        last_dump = cls._next_dump - cls.REPORTING_PERIOD
        cls._next_dump = now + cls.REPORTING_PERIOD

        all_metrics = collector.collect_metrics(*instances)

        for metrics in all_metrics:
            executable = metrics['executable']

            last_reportings = cls._last_reportings[executable]

            for metric, value in metrics.items():
                if not isinstance(value, lifecycle.AbstractGrowingValue):
                    continue

                # compute the difference in value growth to report

                previous_value = last_reportings.get(metric)
                current_value = metrics.get(metric)

                last_reportings[metric] = current_value.latest

                metrics[metric] = current_value.added_content(previous_value)

        cls._reporter.dump_metrics(
            all_metrics, watch_dir=watch_dir, started=cls.STARTED,
            begin=int(last_dump), end=int(now))



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: configuration renderer
#
import os
import stat
import tempfile

import jinja2


def render_configuration(dst_file, template, context):
    if os.path.exists(dst_file):
        os.unlink(dst_file)

    if not context:
        return

    search_path = [
        os.path.join(os.path.dirname(__file__), 'templates')]

    if os.path.sep in template:
        search_path.insert(0, os.path.dirname(template))
        template = os.path.basename(template)

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(search_path),
        trim_blocks=True, lstrip_blocks=True)

    tmpl = env.get_template(template)

    text = tmpl.render(context=context)

    with tempfile.NamedTemporaryFile(
            dir=os.path.dirname(dst_file), mode='w',
            delete=False) as fp:
        fp.write(text)

    os.rename(fp.name, dst_file)
    os.chmod(
        dst_file,
        stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
        stat.S_IRGRP | stat.S_IXGRP |
        stat.S_IROTH | stat.S_IXOTH)



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: REST API metrics server
#
import argparse
import os
import sys

from snmpsim_control_plane.metrics import app
from snmpsim_control_plane.metrics import db
from snmpsim_control_plane.metrics import views  # noqa

DESCRIPTION = """\
SNMP Simulation Control Plane metrics REST API server.

Provides user access to metrics database. The user can request
slices of data by specifying data filtering conditions.

Can be run as a WSGI application.
"""


def parse_args():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        '--recreate-db',
        action='store_true',
        help='DANGER! Running with this flag wipes up REST API server DB! '
             'This switch makes sense only when running this tool for the '
             'first time.')

    parser.add_argument(
        '--config', type=str,
        help='Config file path. Can also be set via environment variable '
             'SNMPSIM_METRICS_CONFIG.')

    parser.add_argument(
        '--interface', type=str,
        help='IP address of the local interface for REST API'
             'server to listen on. Can also be set via config variable '
             'SNMPSIM_METRICS_LISTEN_IP. Default is all local interfaces.')

    parser.add_argument(
        '--port', type=int,
        help='TCP port to bind REST API server to.  Can also be '
             'set via config variable SNMPSIM_METRICS_LISTEN_PORT. '
             'Default is 5001.')

    return parser.parse_args()


def main():
    args = parse_args()

    if args.config:
        os.environ['SNMPSIM_METRICS_CONFIG'] = args.config

    config_file = os.environ.get('SNMPSIM_METRICS_CONFIG')
    if config_file:
        app.config.from_pyfile(config_file)

    if args.interface:
        app.config['SNMPSIM_METRICS_LISTEN_IP'] = args.interface

    if args.port:
        app.config['SNMPSIM_METRICS_LISTEN_PORT'] = args.port

    if args.recreate_db:
        db.drop_all()
        db.create_all()
        return 0

    app.run(host=app.config.get('SNMPSIM_METRICS_LISTEN_IP'),
            port=app.config.get('SNMPSIM_METRICS_LISTEN_PORT'))

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: Process supervisor
#
import argparse
import os
import sys

import snmpsim_control_plane
from snmpsim_control_plane import daemon
from snmpsim_control_plane import log
from snmpsim_control_plane import error
from snmpsim_control_plane.supervisor import manager
from snmpsim_control_plane.supervisor.reporting.manager import ReportingManager


DESCRIPTION = """\
SNMP Simulation Control Plane process supervisor.

Process supervisor watches a directory with potentially changing
executables (e.g. scripts), invokes each and maintains them in running
state for as long as its executable file is in place. On executable
change, the supervisor will restart corresponding processes.

The process supervisor tool is designed to run together with SNMP
simulator configuration management tool.
"""


def parse_args():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        '-v', '--version', action='version',
        version=snmpsim_control_plane.__version__)

    parser.add_argument(
        '--logging-method', type=lambda x: x.split(':'),
        metavar='=<%s[:args]>]' % '|'.join(log.METHODS_MAP),
        default='stderr', help='Logging method')

    parser.add_argument(
        '--log-level', choices=log.LEVELS_MAP,
        type=str, default='info', help='Logging level')

    parser.add_argument(
        '--daemonize', action='store_true',
        help='Disengage from controlling terminal and become a daemon')

    parser.add_argument(
        '--pid-file', metavar='<FILE>', type=str,
        default='/var/run/%s/%s.pid' % (__name__, os.getpid()),
        help='PID file to track supervisor process execution')

    parser.add_argument(
        '--watch-dir', metavar='<DIR>', type=str, required=True,
        help='Location of the executables to herd.')


    parser.add_argument(
        '--reporting-method', type=lambda x: x.split(':'),
        metavar='=<%s[:args]>]' % '|'.join(ReportingManager.REPORTERS),
        default='null', help='SNMP Simulator instance metrics '
                             'reporting method.')

    return parser.parse_args()


def main():
    args = parse_args()

    try:
        log.set_logger(__name__, *args.logging_method, force=True)

        if args.log_level:
            log.set_level(args.log_level)

    except error.ControlPlaneError as exc:
        sys.stderr.write('%s\r\n' % exc)
        return 1

    try:
        ReportingManager.configure(*args.reporting_method)

    except error.ControlPlaneError as exc:
        sys.stderr.write('%s\r\n' % exc)
        return 1

    if args.daemonize:
        try:
            daemon.daemonize(args.pid_file)

        except Exception as exc:
            sys.stderr.write(
                'ERROR: cant daemonize process: %s\r\n' % exc)
            return 1

    manager.manage_executables(args.watch_dir)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator monitor: Flask application
#
import os

from flask import Flask
from flask_marshmallow import Marshmallow
from flask_sqlalchemy import SQLAlchemy

from snmpsim_control_plane.management import config


app = Flask(__name__)

app.url_map.strict_slashes = False

app.config.from_object(config.DefaultConfig)

if 'SNMPSIM_MGMT_CONFIG' in os.environ:
    app.config.from_envvar('SNMPSIM_MGMT_CONFIG')

db = SQLAlchemy(app)
ma = Marshmallow(app)



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: REST API management WSGI app
#
from snmpsim_control_plane.management import app  # noqa
from snmpsim_control_plane.management import views  # noqa



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: Process supervisor metrics
#
import collections

import psutil

from snmpsim_control_plane.supervisor import lifecycle
from snmpsim_control_plane import log


ENDPOINT_MAP = {
    'udp4': 'udpv4',
    'udp6': 'udpv6',
    'tcp4': 'tcpv4',
    'tcp6': 'tcpv6',
}

LIFECYCLE_METRICS = (
    'executable',
    'runtime',
    'exits',
    'changes',
    'console'
)


def collect_metrics(*instances):
    """Collect process metrics

    Example
    -------

    .. code-block::

        {
            'executable': '/path/to/executable',
            'memory': 0,  # memory being used (MB, gauge)
            'cpu': 0,  # consumed cpu time (ms, cumulative)
            'files': 0,  # number of open files (gauge)
            'runtime': 0,  # total time this executable has been running (cumulative)
            'exits': 0,  # number of unexpected exits (cumulative)
            'restarts': 0,  # number of restarts because of changes (cumulative)
            'endpoints': {  # allocated network endpoints (gauge)
                'udpv4': [
                    '127.0.0.1:161',
                    '127.0.0.2:161
                ]
            },
            'console': [
                {
                    'timestamp': {time},
                    'text': '{text}
                }
            ]
        }
    """
    all_metrics = []

    for instance in instances:
        pid = instance['pid']
        if not pid:
            continue

        try:
            process = psutil.Process(pid)

            process_info = process.as_dict()

            endpoints = collections.defaultdict(list)

            for kind in ENDPOINT_MAP:
                for conn in process.connections(kind):
                    endpoints[ENDPOINT_MAP[kind]].append(
                        '%s:%s' % (conn.laddr.ip, conn.laddr.port)
                    )

        except psutil.Error as exc:
            log.error(exc)
            continue

        endpoints = collections.defaultdict(list)

        for kind in ENDPOINT_MAP:
            for conn in process.connections(kind):
                endpoints[ENDPOINT_MAP[kind]].append(
                    '%s:%s' % (conn.laddr.ip, conn.laddr.port)
                )

        metrics = {
            'memory': lifecycle.Gauge(
                process_info['memory_info'].vms // 1024 // 1024),
            'cpu': lifecycle.Counter(
                (process_info['cpu_times'].user +
                 process_info['cpu_times'].system) * 1000),
            'endpoints': endpoints,
            'files': lifecycle.Gauge(process_info['num_fds']),
        }

        metrics.update(
            **{metric: instance[metric] for metric in LIFECYCLE_METRICS})

        all_metrics.append(metrics)

    return all_metrics



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: process management
#
import os
import select
import subprocess
import time

from snmpsim_control_plane import log
from snmpsim_control_plane.supervisor.reporting.manager import ReportingManager
from snmpsim_control_plane.supervisor import lifecycle


POLL_PERIOD = 1

STATE_ADDED = 'added'
STATE_CHANGED = 'changed'
STATE_RUNNING = 'running'
STATE_REMOVED = 'removed'
STATE_DIED = 'died'


def _traverse_dir(top_dir):
    files = []
    entries = os.listdir(top_dir)

    for entry in entries:
        dir_or_file = os.path.join(top_dir, entry)
        if os.path.isdir(dir_or_file):
            files.extend(_traverse_dir(dir_or_file))

        else:
            if os.access(dir_or_file, os.X_OK):
                files.append(dir_or_file)

    return files


def _run_process(fl, fd):
    try:
        return subprocess.Popen(
            [fl], stdout=fd, stderr=fd)

    except Exception as exc:
        log.error('Executable %s failed to start: %s' % (fl, exc))


def _kill_process(leash):
    leash.terminate()

    time.sleep(3)

    if leash.returncode is None:
        log.error(
            'Process %s did not stop gracefully, killing...' % leash.pid)
        leash.kill()
        leash.wait()


def _process_is_running(leash):
    if leash:
        leash.poll()
        return leash.returncode is None


def manage_executables(watch_dir):
    known_instances = {}

    started = int(time.time())

    log.info('Watching directory %s' % watch_dir)

    while True:
        # Collect and log processes output

        rlist = {x['pipe'][0]: x['executable']
                 for x in known_instances.values()
                 if x['state'] == STATE_RUNNING}

        while True:
            try:
                r, w, x = select.select(rlist, [], [], 0.1)

            except Exception as exc:
                log.error(exc)
                break

            if not r:
                break

            timestamp = int(time.time())

            for fd in r:
                executable = rlist[fd]
                instance = known_instances[executable]
                console = instance['console']

                log.msg('Output from process "%s" begins' % executable)

                page_text = os.read(fd, console.MAX_CONSOLE_SIZE)
                page_text = page_text.decode(errors='ignore')

                console.add(page_text, timestamp)

                log.msg(page_text)
                log.msg('Output from process "%s" ends' % executable)

        # Watch executables

        existing_files = set()

        try:
            files = _traverse_dir(watch_dir)

        except Exception as exc:
            log.error('Directory %s traversal failure: %s' % (watch_dir, exc))
            time.sleep(10)
            continue

        for fl in files:
            instance = known_instances.get(fl)

            stat = os.stat(fl).st_mtime

            if not instance:
                instance = {
                    'pid': 0,
                    'executable': fl,
                    'file_info': stat,
                    'leash': None,
                    'pipe': (None, None),
                    'state': STATE_ADDED,
                    'created': time.time(),
                    'started': None,
                    'stopped': None,
                    'runtime': lifecycle.Counter(0),
                    'changes': lifecycle.Counter(0),
                    'exits': lifecycle.Counter(0),
                    'console': lifecycle.ConsoleLog(),
                }
                known_instances[fl] = instance

                log.info('Start tracking executable %s' % fl)

            pid = instance['leash'].pid if instance['leash'] else '?'

            if instance['file_info'] != stat:
                instance['file_info'] = stat
                instance['state'] = STATE_CHANGED
                instance['changes'] += 1

                log.info('Existing executable %s (PID %s) has '
                         'changed' % (fl, pid))

            if instance['state'] == STATE_RUNNING:
                executable = instance['leash']

                executable.poll()

                if executable.returncode is not None:
                    instance['state'] = STATE_DIED
                    instance['stopped'] = time.time()
                    instance['exits'] += 1

                    uptime = int(
                        time.time() - instance['started'] or time.time())

                    log.info(
                        'Executable %s (PID %s) has died '
                        '(rc=%s), uptime %s' % (fl, pid, executable.returncode,
                                                uptime))

            existing_files.add(fl)

        removed_files = set(known_instances) - existing_files

        for fl in removed_files:
            instance = known_instances[fl]
            instance['state'] = STATE_REMOVED
            instance['changes'] += 1

            log.info(
                'Existing executable %s (PID %s) has been '
                'removed' % (fl, instance['pid']))

        for fl, instance in tuple(known_instances.items()):
            state = instance['state']

            if state in (STATE_ADDED, STATE_DIED):
                if state == STATE_DIED:
                    r, w = instance['pipe']

                    try:
                        os.close(r)
                        os.close(w)

                    except OSError as exc:
                        log.error(exc)

                r, w = os.pipe()

                leash = _run_process(fl, w)

                instance['leash'] = leash
                instance['pipe'] = r, w

                if leash:
                    instance['state'] = STATE_RUNNING
                    instance['started'] = time.time()
                    instance['pid'] = leash.pid

                    log.info(
                        'Executable %s (PID %s) has been '
                        'started' % (fl, leash.pid))

            elif state in (STATE_CHANGED, STATE_REMOVED):
                leash = instance['leash']

                if leash:
                    _kill_process(leash)

                    log.info(
                        'Executable %s (PID %s) has been '
                        'stopped' % (fl, leash.pid))

                r, w = instance['pipe']
                if r:
                    try:
                        os.close(r)
                        os.close(w)

                    except OSError as exc:
                        log.error(exc)

                if state == STATE_CHANGED:
                    instance['state'] = STATE_DIED

                else:
                    known_instances.pop(fl)

                    log.info(
                        'Stopped tracking executable %s' % fl)

            elif state == STATE_RUNNING:
                leash = instance['leash']
                if _process_is_running(leash):
                    now = time.time()
                    instance['runtime'] = lifecycle.Counter(
                        now - instance['created'])

                else:
                    instance['state'] = STATE_DIED
                    instance['exits'] += 1

                    log.info('Executable %s (PID %s) has '
                             'died' % (fl, leash.pid))

        ReportingManager.process_metrics(
            watch_dir, *known_instances.values())

        time.sleep(POLL_PERIOD)



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: turn ORM data into a tree of dicts
#
from sqlalchemy import and_
from sqlalchemy import inspect

from snmpsim_control_plane.management import models


def object_as_dict(obj):
    return {
        c.key: getattr(obj, c.key)
        for c in inspect(obj).mapper.column_attrs
    }


def to_dict():
    labs = []

    query = (
        models.Lab
        .query
        .filter_by(power='on').all()
    )

    for orm_lab in query:
        agents = []

        lab = object_as_dict(orm_lab)
        lab.update(agents=agents)
        labs.append(lab)

        query = (
            models.Agent
            .query
            .join(models.LabAgent,
                  and_(models.LabAgent.agent_id == models.Agent.id,
                       models.LabAgent.lab_id == orm_lab.id))
            .all()
        )

        for orm_agent in query:
            engines = []
            selectors = []

            agent = object_as_dict(orm_agent)
            agent.update(engines=engines, selectors=selectors)
            agents.append(agent)

            query = (
                models.Engine
                .query
                .join(models.AgentEngine,
                      and_(models.AgentEngine.engine_id == models.Engine.id,
                           models.AgentEngine.agent_id == orm_agent.id))
                .all()
            )

            for orm_engine in query:
                users = []
                endpoints = []

                engine = object_as_dict(orm_engine)
                engine.update(users=users, endpoints=endpoints)
                engines.append(engine)

                query = (
                    models.User
                    .query
                    .join(models.EngineUser,
                          and_(models.EngineUser.user_id == models.User.id,
                               models.EngineUser.engine_id == orm_engine.id))
                    .all()
                )

                for orm_user in query:
                    user = object_as_dict(orm_user)
                    users.append(user)

                query = (
                    models.Endpoint
                    .query
                    .join(models.EngineEndpoint,
                          and_(models.EngineEndpoint.endpoint_id == models.Endpoint.id,
                               models.EngineEndpoint.engine_id == orm_engine.id))
                    .all()
                )

                for orm_endpoint in query:
                    endpoint = object_as_dict(orm_endpoint)
                    endpoints.append(endpoint)

            query = (
                models.Selector
                .query
                .join(models.AgentSelector,
                      and_(models.AgentSelector.selector_id == models.Selector.id,
                           models.AgentSelector.agent_id == orm_agent.id))
                .all()
            )

            for orm_selector in query:
                selector = object_as_dict(orm_selector)
                selectors.append(selector)

    context = {}

    if labs:
        context.update(labs=labs)

    return context




#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: REST API views
#
import os
import tempfile
from functools import wraps

import flask
from sqlalchemy import func
from werkzeug import exceptions

from snmpsim_control_plane import error
from snmpsim_control_plane.management import app
from snmpsim_control_plane.management import db
from snmpsim_control_plane.management import models
from snmpsim_control_plane.management import recording
from snmpsim_control_plane.management import schemas
from snmpsim_control_plane.management.exporters import builder
from snmpsim_control_plane.management.exporters import renderer

PREFIX = '/snmpsim/mgmt/v1'
TARGET_CONFIG = 'snmpsim-run-labs.sh'


def render_config(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        response = f(*args, **kwargs)

        context = builder.to_dict()

        template = app.config['SNMPSIM_MGMT_TEMPLATE']
        dst = os.path.join(
            app.config['SNMPSIM_MGMT_DESTINATION'],
            TARGET_CONFIG)

        renderer.render_configuration(dst, template, context)

        return response

    return decorated_function


@app.errorhandler(exceptions.HTTPException)
def flask_exception_handler(exc):
    app.logger.error(exc)
    err = {
        'status': exc.code,
        'message': exc.description
    }
    response = flask.jsonify(err)
    response.status_code = exc.code
    return response


@app.errorhandler(Exception)
def all_exception_handler(exc):
    app.logger.error(exc)
    err = {
        'status': 400,
        'message': getattr(exc, 'message', str(exc))
    }
    response = flask.jsonify(err)
    response.status_code = 400
    return response


def cleanup_recordings(f):
    @wraps(f)
    def rm_empty_dirs(top_dir, delete=False):
        entries = os.listdir(top_dir)

        for entry in entries:
            sub_dir = os.path.join(top_dir, entry)
            if not os.path.isdir(sub_dir):
                continue

            rm_empty_dirs(sub_dir, delete=True)

        if delete:
            try:
                os.rmdir(top_dir)

            except OSError:
                return

    def decorated_function(*args, **kwargs):
        response = f(*args, **kwargs)

        rm_empty_dirs(app.config['SNMPSIM_MGMT_DATAROOT'])

        return response

    return decorated_function


def search_model(model, query):

    known_columns = model.__table__.columns.keys()

    for search_column in flask.request.args:
        if search_column not in known_columns:
            raise exceptions.NotFound(
                'Search term %s is not supported' % search_column)

        search_terms = flask.request.args.getlist(search_column)
        if search_terms:
            search_terms = [term.lower() for term in search_terms]
            query = query.filter(
                func.lower(getattr(model, search_column)).in_(search_terms))

    return query


@app.route(PREFIX + '/endpoints')
def show_endpoints():
    endpoints_query = (
        models.Endpoint
        .query
        .outerjoin(models.EngineEndpoint)
        .outerjoin(models.Engine))

    endpoints_query = search_model(models.Endpoint, endpoints_query)

    schema = schemas.EndpointSchema(many=True)
    return schema.jsonify(endpoints_query.all())


@app.route(PREFIX + '/endpoints/<id>', methods=['GET'])
def show_endpoint(id):
    endpoint = (
        models.Endpoint
        .query
        .filter_by(id=id)
        .outerjoin(models.EngineEndpoint)
        .outerjoin(models.Engine)
        .first())

    if not endpoint:
        raise exceptions.NotFound('Endpoint not found')

    schema = schemas.EndpointSchema()
    return schema.jsonify(endpoint), 200


@app.route(PREFIX + '/endpoints', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/endpoint', methods=['POST'])
@render_config
def new_endpoint(tag_id=None):
    req = flask.request.json

    endpoint = models.Endpoint(**req)
    db.session.add(endpoint)

    if tag_id:
        db.session.flush()
        tag_endpoint = models.TagEndpoint(
            tag_id=tag_id, endpoint_id=endpoint.id)
        db.session.add(tag_endpoint)

    db.session.commit()

    schema = schemas.EndpointSchema()
    return schema.jsonify(endpoint), 201


@app.route(PREFIX + '/endpoints/<id>', methods=['DELETE'])
@render_config
def del_endpoint(id):
    endpoint = (
        models.Endpoint
        .query
        .filter_by(id=id)
        .first())

    if not endpoint:
        raise exceptions.NotFound('Endpoint not found')

    db.session.delete(endpoint)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/users')
def show_users():
    users_query = (
        models.User
        .query)

    users_query = search_model(models.User, users_query)

    schema = schemas.UserSchema(many=True)
    return schema.jsonify(users_query.all()), 200


@app.route(PREFIX + '/users/<id>', methods=['GET'])
def show_user(id):
    user = (
        models.User
        .query
        .filter_by(id=id)
        .first())

    if not user:
        raise exceptions.NotFound('User not found')

    schema = schemas.UserSchema()
    return schema.jsonify(user), 200


@app.route(PREFIX + '/users', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/user', methods=['POST'])
@render_config
def new_user(tag_id=None):
    req = flask.request.json

    user = models.User(**req)
    db.session.add(user)

    if tag_id:
        db.session.flush()
        tag_user = models.TagUser(
            tag_id=tag_id, user_id=user.id)
        db.session.add(tag_user)

    db.session.commit()

    schema = schemas.UserSchema()
    return schema.jsonify(user), 201


@app.route(PREFIX + '/users/<id>', methods=['DELETE'])
@render_config
def del_user(id):
    user = (
        models.User
        .query
        .filter_by(id=id)
        .first())

    if not user:
        raise exceptions.NotFound('User not found')

    db.session.delete(user)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/engines')
def show_engines():
    engines_query = (
        models.Engine
        .query
        .outerjoin(models.EngineUser)
        .outerjoin(models.User))

    engines_query = search_model(models.Engine, engines_query)

    schema = schemas.EngineSchema(many=True)
    return schema.jsonify(engines_query.all()), 200


@app.route(PREFIX + '/engines/<id>', methods=['GET'])
def show_engine(id):
    engine = (
        models.Engine
        .query
        .filter_by(id=id)
        .outerjoin(models.EngineUser)
        .outerjoin(models.User)
        .first())

    if not engine:
        return flask.make_response('Engine not found', 404)

    schema = schemas.EngineSchema()
    return schema.jsonify(engine), 200


@app.route(PREFIX + '/engines', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/engine', methods=['POST'])
@render_config
def new_engine(tag_id=None):
    req = flask.request.json

    engine = models.Engine(**req)
    db.session.add(engine)

    if tag_id:
        db.session.flush()
        tag_engine = models.TagEngine(
            tag_id=tag_id, engine_id=engine.id)
        db.session.add(tag_engine)

    db.session.commit()

    schema = schemas.EngineSchema()
    return schema.jsonify(engine), 201


@app.route(PREFIX + '/engines/<id>', methods=['DELETE'])
@render_config
def del_engine(id):
    engine = (
        models
        .Engine
        .query
        .filter_by(id=id)
        .first())

    if not engine:
        raise exceptions.NotFound('Engine not found')

    db.session.delete(engine)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/engines/<id>/user/<user_id>', methods=['PUT'])
@render_config
def add_engine_user(id, user_id):
    engine_user = models.EngineUser(user_id=user_id, engine_id=id)

    db.session.add(engine_user)
    db.session.commit()

    engine = models.Engine.query.filter_by(id=id).first()
    schema = schemas.EngineSchema()
    return schema.jsonify(engine), 200


@app.route(PREFIX + '/engines/<id>/user/<user_id>', methods=['DELETE'])
@render_config
def del_engine_user(id, user_id):
    engine_user = (
        models
        .EngineUser
        .query
        .filter_by(engine_id=id, user_id=user_id)
        .first())

    if not engine_user:
        raise exceptions.NotFound('Engine <-> user binding not found')

    db.session.delete(engine_user)
    db.session.commit()

    engine = models.Engine.query.filter_by(id=id).first()
    schema = schemas.EngineSchema()
    return schema.jsonify(engine), 204


@app.route(PREFIX + '/engines/<id>/endpoint/<endpoint_id>', methods=['PUT'])
@render_config
def add_engine_endpoint(id, endpoint_id):
    engine_endpoint = models.EngineEndpoint(
        endpoint_id=endpoint_id, engine_id=id)

    db.session.add(engine_endpoint)
    db.session.commit()

    engine = (
        models
        .Engine
        .query
        .filter_by(id=id).first())

    schema = schemas.EngineSchema()
    return schema.jsonify(engine), 200


@app.route(PREFIX + '/engines/<id>/endpoint/<endpoint_id>', methods=['DELETE'])
@render_config
def del_engine_endpoint(id, endpoint_id):
    engine_endpoint = (
        models
        .EngineEndpoint
        .query
        .filter_by(engine_id=id, endpoint_id=endpoint_id).first())

    if not engine_endpoint:
        raise exceptions.NotFound('Engine <-> endpoint binding not found')

    db.session.delete(engine_endpoint)
    db.session.commit()

    engine = models.Engine.query.filter_by(id=id).first()
    schema = schemas.EngineSchema()

    return schema.jsonify(engine), 204


@app.route(PREFIX + '/agents')
def show_agents():
    agents_query = (
        models.Agent
        .query
        .outerjoin(models.AgentEngine)
        .outerjoin(models.Engine))

    agents_query = search_model(models.Agent, agents_query)

    schema = schemas.AgentSchema(many=True)
    return schema.jsonify(agents_query.all()), 200


@app.route(PREFIX + '/agents/<id>', methods=['GET'])
def show_agent(id):
    agent = (
        models.Agent
        .query
        .filter_by(id=id)
        .outerjoin(models.AgentEngine)
        .outerjoin(models.Engine)
        .first())

    if not agent:
        raise exceptions.NotFound('Agent not found')

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 200


@app.route(PREFIX + '/agents', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/agent', methods=['POST'])
@render_config
def new_agent(tag_id=None):
    req = flask.request.json

    data_dir = req.get('data_dir')
    if data_dir:
        data_dir = os.path.join(
            app.config['SNMPSIM_MGMT_DATAROOT'],
            data_dir)

        data_dir = os.path.abspath(data_dir)

        if not data_dir.startswith(
                os.path.abspath(app.config['SNMPSIM_MGMT_DATAROOT'])):
            raise error.ControlPlaneError(
                'Data directory outside of data root: %s' % data_dir)

        req.update(data_dir=data_dir)

    agent = models.Agent(**req)
    db.session.add(agent)

    if tag_id:
        db.session.flush()
        tag_agent = models.TagAgent(
            tag_id=tag_id, agent_id=agent.id)
        db.session.add(tag_agent)

    db.session.commit()

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 201


@app.route(PREFIX + '/agents/<id>', methods=['DELETE'])
@render_config
def del_agent(id):
    agent = (
        models
        .Agent
        .query
        .filter_by(id=id)
        .first())

    if not agent:
        raise exceptions.NotFound('Agent not found')

    db.session.delete(agent)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/agents/<id>/engine/<engine_id>', methods=['PUT'])
@render_config
def add_agent_engine(id, engine_id):
    agent_engine = (
        models
        .AgentEngine(agent_id=id, engine_id=engine_id))

    db.session.add(agent_engine)
    db.session.commit()

    agent = (
        models
        .Agent
        .query
        .filter_by(id=id).first())

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 200


@app.route(PREFIX + '/agents/<id>/engine/<engine_id>', methods=['DELETE'])
@render_config
def del_agent_engine(id, engine_id):
    agent_engine = (
        models
        .AgentEngine
        .query
        .filter_by(engine_id=engine_id, agent_id=id).first())

    if not agent_engine:
        raise exceptions.NotFound('Agent <-> engine binding not found')

    db.session.delete(agent_engine)
    db.session.commit()

    agent = (
        models
        .Agent
        .query
        .filter_by(id=id).first())

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 204


@app.route(PREFIX + '/selectors')
def show_selectors():
    selectors_query = (
        models.Selector
        .query
        .outerjoin(models.AgentSelector)
        .outerjoin(models.Agent))

    selectors_query = search_model(models.Selector, selectors_query)

    schema = schemas.SelectorSchema(many=True)
    return schema.jsonify(selectors_query.all()), 200


@app.route(PREFIX + '/selectors/<id>', methods=['GET'])
def show_selector(id):
    selector_query = (
        models.Selector
        .query
        .filter_by(id=id)
        .outerjoin(models.AgentSelector)
        .outerjoin(models.Agent)
        .first())

    if not selector_query:
        raise exceptions.NotFound('Selector not found')

    schema = schemas.SelectorSchema()
    return schema.jsonify(selector_query), 200


@app.route(PREFIX + '/selectors', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/selector', methods=['POST'])
@render_config
def new_selector(tag_id=None):
    req = flask.request.json

    selector = models.Selector(**req)
    db.session.add(selector)

    if tag_id:
        db.session.flush()
        tag_selector = models.TagSelector(
            tag_id=tag_id, selector_id=selector.id)
        db.session.add(tag_selector)

    db.session.commit()

    schema = schemas.SelectorSchema()
    return schema.jsonify(selector), 201


@app.route(PREFIX + '/selectors/<id>', methods=['DELETE'])
@render_config
def del_selector(id):
    selector = (
        models.Selector
        .query
        .filter_by(id=id)
        .first())

    if not selector:
        raise exceptions.NotFound('Selector not found')

    db.session.delete(selector)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/agents/<id>/selector/<selector_id>/<order>',
           methods=['PUT'])
@render_config
def add_agent_selector(id, selector_id, order):
    agent_selector = (
        models
        .AgentSelector(agent_id=id, selector_id=selector_id, order=order))

    db.session.add(agent_selector)
    db.session.commit()

    agent = (
        models
        .Agent
        .query
        .filter_by(id=id).first())

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 200


@app.route(PREFIX + '/agents/<id>/selector/<selector_id>', methods=['DELETE'])
@render_config
def del_agent_selector(id, selector_id):
    agent_selector = (
        models
        .AgentSelector
        .query
        .filter_by(selector_id=selector_id, agent_id=id).first())

    if not agent_selector:
        raise exceptions.NotFound('Agent <-> selector binding not found')

    db.session.delete(agent_selector)
    db.session.commit()

    agent = (
        models
        .Agent
        .query
        .filter_by(id=id).first())

    schema = schemas.AgentSchema()
    return schema.jsonify(agent), 204


@app.route(PREFIX + '/recordings')
def show_recordings():
    try:
        recordings = recording.list_recordings(
            app.config['SNMPSIM_MGMT_DATAROOT'])

    except error.ControlPlaneError:
        raise exceptions.NotFound('Recording not found')

    schema = schemas.RecordingSchema(many=True)
    return schema.jsonify(recordings), 200


@app.route(PREFIX + '/recordings/<path:path>', methods=['GET'])
def show_recording(path):
    try:
        directory, file = recording.get_recording(
            app.config['SNMPSIM_MGMT_DATAROOT'], path, exists=True)

    except error.ControlPlaneError:
        raise exceptions.NotFound('Recording not found')

    return flask.send_from_directory(directory, file)


@app.route(PREFIX + '/recordings/<path:path>', methods=['POST', 'PUT'])
def new_recording(path):
    recording_type = recording.get_recording_type(path)
    if not recording_type:
        raise error.ControlPlaneError('Unknown recording type')

    can_exist = flask.request.method == 'POST'

    try:
        directory, file = recording.get_recording(
            app.config['SNMPSIM_MGMT_DATAROOT'], path,
            not_exists=can_exist, ensure_path=True)

    except error.ControlPlaneError:
        raise exceptions.Conflict('Bad recording path (is it already exists?)')

    # TODO: is it a memory hog when .snmprec file is large?
    with tempfile.NamedTemporaryFile(dir=directory, delete=False) as fl:
        fl.write(flask.request.data)

    os.rename(fl.name, os.path.join(directory, file))

    return flask.Response(status=204)


@app.route(PREFIX + '/recordings/<path:path>', methods=['DELETE'])
@cleanup_recordings
def del_recording(path):
    try:
        directory, file = recording.get_recording(
            app.config['SNMPSIM_MGMT_DATAROOT'], path, exists=True)

    except error.ControlPlaneError:
        raise exceptions.NotFound('Recording not found')

    os.unlink(os.path.join(directory, file))

    return flask.Response(status=204)


@app.route(PREFIX + '/labs')
def show_labs():
    labs_query = (
        models.Lab
        .query
        .outerjoin(models.LabAgent)
        .outerjoin(models.Agent))

    labs_query = search_model(models.Lab, labs_query)

    schema = schemas.LabSchema(many=True)
    return schema.jsonify(labs_query.all()), 200


@app.route(PREFIX + '/labs/<id>', methods=['GET'])
def show_lab(id):
    lab = (
        models.Lab
        .query
        .filter_by(id=id)
        .outerjoin(models.LabAgent)
        .outerjoin(models.Agent)
        .first())

    if not lab:
        raise exceptions.NotFound('Lab not found')

    schema = schemas.LabSchema()
    return schema.jsonify(lab), 200


@app.route(PREFIX + '/labs', methods=['POST'])
@app.route(PREFIX + '/tags/<tag_id>/lab', methods=['POST'])
@render_config
def new_lab(tag_id=None):
    req = flask.request.json

    lab = models.Lab(**req)
    db.session.add(lab)

    if tag_id:
        db.session.flush()
        tag_lab = models.TagLab(
            tag_id=tag_id, lab_id=lab.id)
        db.session.add(tag_lab)

    db.session.commit()

    schema = schemas.LabSchema()
    return schema.jsonify(lab), 201


@app.route(PREFIX + '/labs/<id>', methods=['DELETE'])
@render_config
def del_lab(id):
    lab = (
        models
        .Lab
        .query
        .filter_by(id=id)
        .first())

    if not lab:
        raise exceptions.NotFound('Lab not found')

    db.session.delete(lab)
    db.session.commit()

    return flask.Response(status=204)


@app.route(PREFIX + '/labs/<id>/agent/<agent_id>', methods=['PUT'])
@render_config
def add_lab_agent(id, agent_id):
    lab_agent = models.LabAgent(agent_id=agent_id, lab_id=id)

    db.session.add(lab_agent)
    db.session.commit()

    lab = models.Lab.query.filter_by(id=id).first()

    schema = schemas.LabSchema()
    return schema.jsonify(lab), 200


@app.route(PREFIX + '/labs/<id>/agent/<agent_id>', methods=['DELETE'])
@render_config
def del_lab_agent(id, agent_id):
    lab_agent = (
        models
        .LabAgent
        .query
        .filter_by(agent_id=agent_id, lab_id=id)
        .first())

    if not lab_agent:
        raise exceptions.NotFound('Lab <-> agent binding not found')

    db.session.delete(lab_agent)
    db.session.commit()

    engine = models.Lab.query.filter_by(id=id).first()
    schema = schemas.LabSchema()
    return schema.jsonify(engine), 204


@app.route(PREFIX + '/labs/<id>/power/<state>', methods=['PUT'])
@render_config
def change_lab_power(id, state):
    lab = (
        models.Lab
        .query
        .filter_by(id=id)
        .first())

    if not lab:
        raise exceptions.NotFound('Lab not found')

    lab.power = state.lower()

    db.session.commit()

    lab = models.Lab.query.filter_by(id=id).first()

    schema = schemas.LabSchema()
    return schema.jsonify(lab), 200


def make_tags_query(tag_id=None):
    tags_query = (
        models.Tag
        .query)

    if tag_id is not None:
        tags_query = tags_query.filter_by(id=tag_id)

    tags_query = (
        tags_query
        .outerjoin(models.TagEndpoint)
        .outerjoin(models.Endpoint)
        .outerjoin(models.TagUser)
        .outerjoin(models.User)
        .outerjoin(models.TagEngine)
        .outerjoin(models.Engine)
        .outerjoin(models.TagSelector)
        .outerjoin(models.Selector)
        .outerjoin(models.TagAgent)
        .outerjoin(models.Agent)
        .outerjoin(models.TagLab)
        .outerjoin(models.Lab))

    return tags_query


@app.route(PREFIX + '/tags')
def show_tags():
    tags_query = make_tags_query()

    tags_query = search_model(models.Tag, tags_query)

    schema = schemas.TagSchema(many=True)
    return schema.jsonify(tags_query.all()), 200


@app.route(PREFIX + '/tags/<id>', methods=['GET'])
def show_tag(id):
    tags_query = make_tags_query(id)

    tag = tags_query.first()
    if not tag:
        raise exceptions.NotFound('Tag not found')

    schema = schemas.TagSchema()
    return schema.jsonify(tag), 200


@app.route(PREFIX + '/tags', methods=['POST'])
@render_config
def new_tag():
    req = flask.request.json

    tag = models.Tag(**req)
    db.session.add(tag)
    db.session.commit()

    tags_query = make_tags_query(tag.id)

    schema = schemas.TagSchema()
    return schema.jsonify(tags_query.first()), 201


@app.route(PREFIX + '/tags/<id>', methods=['DELETE'])
@render_config
def del_tag(id):
    tag_query = (
        models.Tag
        .query
        .filter_by(id=id))

    tag = tag_query.first()
    if not tag:
        raise exceptions.NotFound('Tag not found')

    db.session.delete(tag)
    db.session.commit()

    return flask.Response(status=204)


ENTITY_ADD_MAP = {
    'endpoint': lambda x, y: models.TagEndpoint(tag_id=x, endpoint_id=y),
    'user': lambda x, y: models.TagUser(tag_id=x, user_id=y),
    'engine': lambda x, y: models.TagEngine(tag_id=x, engine_id=y),
    'selector': lambda x, y: models.TagSelector(tag_id=x, selector_id=y),
    'agent': lambda x, y: models.TagAgent(tag_id=x, agent_id=y),
    'lab': lambda x, y: models.TagLab(tag_id=x, lab_id=y),
}


ENTITY_DEL_MAP = {
    'endpoint': lambda x, y: (
        models.TagEndpoint
        .query
        .filter_by(tag_id=x, endpoint_id=y)
    ),
    'user': lambda x, y: (
        models.TagUser
        .query
        .filter_by(tag_id=x, user_id=y)
    ),
    'engine': lambda x, y: (
        models.TagEngine
        .query
        .filter_by(tag_id=x, engine_id=y)
    ),
    'selector': lambda x, y: (
        models.TagSelector
        .query
        .filter_by(tag_id=x, selector_id=y)
    ),
    'agent': lambda x, y: (
        models
        .TagAgent
        .query
        .filter_by(tag_id=x, agent_id=y)
    ),
    'lab': lambda x, y: (
        models
        .TagLab
        .query
        .filter_by(tag_id=x, lab_id=y)
    ),
}


@app.route(PREFIX + '/tags/<id>/<entity>/<entity_id>', methods=['PUT'])
@render_config
def add_tag_entity(id, entity, entity_id):
    try:
        tag_entity = ENTITY_ADD_MAP[entity]

    except KeyError:
        raise exceptions.NotFound('Entity %s not found' % entity)

    tag_entity = tag_entity(id, entity_id)

    db.session.add(tag_entity)
    db.session.commit()

    tags_query = make_tags_query(id)

    schema = schemas.TagSchema()
    return schema.jsonify(tags_query.first()), 200


@app.route(PREFIX + '/tags/<id>/<entity>/<entity_id>', methods=['DELETE'])
@render_config
def del_tag_entity(id, entity, entity_id):
    try:
        entity_query = ENTITY_DEL_MAP[entity]

    except KeyError:
        raise exceptions.NotFound('Entity %s not found' % entity)

    entity_query = entity_query(id, entity_id)

    entity_instance = entity_query.first()

    if not entity_instance:
        raise exceptions.NotFound('Tag <-> %s binding not found' % entity)

    db.session.delete(entity_instance)
    db.session.commit()

    entity_query = make_tags_query(id)

    schema = schemas.TagSchema()
    return schema.jsonify(entity_query.first()), 200


DELETABLE_ENTITIES = (
    'endpoints', 'users', 'engines', 'selectors', 'agents', 'labs'
)


@app.route(PREFIX + '/tags/<id>/objects', methods=['DELETE'])
@render_config
def del_tagged_objects(id):
    tags_query = make_tags_query(id)

    tag = tags_query.first()
    if not tag:
        raise exceptions.NotFound('Tag with ID %s not found' % id)

    for entity_name in DELETABLE_ENTITIES:
        for entity in getattr(tag, entity_name):
            db.session.delete(entity)

    db.session.commit()

    entity_query = make_tags_query(id)

    schema = schemas.TagSchema()
    return schema.jsonify(entity_query.first()), 200



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: Process supervisor metrics
#


class BaseReporter(object):
    """Maintain activity metrics.
    """
    def __init__(self, *args, **kwargs):
        pass

    def dump_metrics(self, metrics, watch_dir=None,
                     started=None, begin=None, end=None):
        """Dump metrics in a reporter-specific way.
        """

    def __str__(self):
        return self.__class__.__name__



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: REST API management server
#


class DefaultConfig(object):
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory'
    SQLALCHEMY_ECHO = False

    DEBUG = False

    SNMPSIM_MGMT_LISTEN_IP = '127.0.0.1'
    SNMPSIM_MGMT_LISTEN_PORT = 5000
    SNMPSIM_MGMT_SSL_CERT = None
    SNMPSIM_MGMT_SSL_KEY = None
    SNMPSIM_MGMT_DATAROOT = '/usr/share/snmpsim/data'
    SNMPSIM_MGMT_TEMPLATE = 'snmpsim-command-responder.j2'
    SNMPSIM_MGMT_DESTINATION = '.'



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: process lifecycle support
#


class AbstractGrowingValue(object):
    """Interface for computing how much a value has grown.
    """

    @property
    def latest(self):
        raise NotImplementedError()

    def added_content(self, relative_to=None):
        raise NotImplementedError()


class Gauge(AbstractGrowingValue, int):
    """Gauge integer number.

    Gauge objects can increase and decrease in value over time.
    """
    @property
    def latest(self):
        return 0

    def added_content(self, relative_to=None):
        return self

    def __add__(self, other):
        return Gauge(int(self) + other)

    def __sub__(self, other):
        return Gauge(int(self) - other)

    __iadd__ = __add__
    __isub__ = __sub__


class Counter(AbstractGrowingValue, int):
    """Integer counter.

    Counter object can only increase in value over time.
    """
    @property
    def latest(self):
        return self

    def added_content(self, relative_to=None):
        return max(0, self - (relative_to or 0))

    def __add__(self, other):
        return Counter(int(self) + other)

    __iadd__ = __add__


class ConsoleLog(AbstractGrowingValue):
    """Paged process console log.

    Text pages can be added at the end of the log, and automatically
    expire.
    """
    MAX_CONSOLES = 50
    MAX_CONSOLE_SIZE = 80 * 24  # tribute to VT100

    def __init__(self, first_page=0):
        self._first_page = first_page
        self._last_page = first_page
        self._text = {}
        self._timestamps = {}

    @property
    def first_page(self):
        return self._first_page

    @property
    def last_page(self):
        return self._last_page - 1

    def add(self, text, timestamp):
        self._text[self._last_page] = text
        self._timestamps[self._last_page] = timestamp
        self._last_page += 1

        if len(self._text) > self.MAX_CONSOLES:
            self._text.pop(self._first_page)
            self._timestamps.pop(self._first_page)
            self._first_page += 1

    def text(self, page):
        return self._text.get(page, '')

    def timestamp(self, page):
        return self._timestamps.get(page, '')

    @property
    def latest(self):
        return self.last_page

    def added_content(self, relative_to=None):
        first_page = relative_to or 0
        console = self.__class__(first_page=first_page)

        for page in range(first_page, self.last_page):
            console.add(self.text(page), self.timestamp(page))

        return console




#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: metrics importing tool
#
import argparse
import sys
import os

import snmpsim_control_plane
from snmpsim_control_plane import daemon
from snmpsim_control_plane import error
from snmpsim_control_plane import log
from snmpsim_control_plane.metrics import app
from snmpsim_control_plane.metrics import db
from snmpsim_control_plane.metrics import reader
from snmpsim_control_plane.metrics import models  # noqa

DESCRIPTION = """\
SNMP Simulation Control Plane metrics importer.

Watches given directory for metrics JSON files to appear,
load them into ORM and remove once processed.
"""


class DefaultConfig(object):
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory'
    SQLALCHEMY_ECHO = False

    DEBUG = False


def parse_args():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        '-v', '--version', action='version',
        version=snmpsim_control_plane.__version__)

    parser.add_argument(
        '--recreate-db',
        action='store_true',
        help='DANGER! Running with this flag wipes up metrics DB! '
             'This switch makes sense only when running this tool for the '
             'first time.')

    parser.add_argument(
        '--config', type=str,
        help='Config file path. Can also be set via environment variable '
             'SNMPSIM_METRICS_CONFIG.')

    parser.add_argument(
        '--logging-method', type=lambda x: x.split(':'),
        metavar='=<%s[:args]>]' % '|'.join(log.METHODS_MAP),
        default='stderr', help='Logging method')

    parser.add_argument(
        '--log-level', choices=log.LEVELS_MAP,
        type=str, default='info', help='Logging level')

    parser.add_argument(
        '--daemonize', action='store_true',
        help='Disengage from controlling terminal and become a daemon')

    parser.add_argument(
        '--pid-file', metavar='<FILE>', type=str,
        default='/var/run/%s/%s.pid' % (__name__, os.getpid()),
        help='PID file to track daemon process execution')

    parser.add_argument(
        '--watch-dir', metavar='<DIR>', type=str,
        help='Location of the metrics JSON files to import and remove.')

    return parser.parse_args()


def main():
    args = parse_args()

    app.config.from_object(DefaultConfig)

    if args.config:
        app.config.from_pyfile(args.config)

    if args.recreate_db:
        db.drop_all()
        db.create_all()
        return 0

    if not args.watch_dir:
        sys.stderr.write('ERROR: --watch-dir must be specified\r\n')
        return 1

    try:
        log.set_logger(__name__, *args.logging_method, force=True)

        if args.log_level:
            log.set_level(args.log_level)

    except error.ControlPlaneError as exc:
        sys.stderr.write('%s\r\n' % exc)
        return 1

    if args.daemonize:
        try:
            daemon.daemonize(args.pid_file)

        except Exception as exc:
            sys.stderr.write(
                'ERROR: cant daemonize process: %s\r\n' % exc)
            return 1

    reader.watch_metrics(args.watch_dir)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: metrics files reader
#
import os
import json
import time

from snmpsim_control_plane import log
from snmpsim_control_plane.metrics import manager

POLL_PERIOD = 10


def _traverse_dir(dir):
    files = []
    entries = os.listdir(dir)

    for entry in entries:
        dir_or_file = os.path.join(dir, entry)
        if os.path.isdir(dir_or_file):
            files.extend(_traverse_dir(dir_or_file))

        else:
            files.append(dir_or_file)

    return files


def watch_metrics(watch_dir):

    log.info('Watching directory %s' % watch_dir)

    while True:

        try:
            files = _traverse_dir(watch_dir)

        except Exception as exc:
            log.error('Directory %s traversal failure: %s' % (watch_dir, exc))
            time.sleep(10)
            continue

        for filename in files:

            log.info('Processing %s' % filename)

            try:
                with open(filename) as fl:
                    jsondoc = json.loads(fl.read())

            except Exception as exc:
                log.error('Error reading file %s: %s' % (filename, exc))
                continue

            finally:
                os.unlink(filename)

            try:
                manager.import_metrics(jsondoc)

            except Exception as exc:
                log.error('Error processing file %s: %s' % (filename, exc))
                continue

        time.sleep(POLL_PERIOD)



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: REST API view schemas
#

from snmpsim_control_plane.management import ma
from snmpsim_control_plane.management import models


class MinimalTagSchema(ma.ModelSchema):
    class Meta:
        model = models.Tag
        fields = ('id', 'name', 'description', '_links')

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_tag', id='<id>'),
        'collection': ma.URLFor('show_tags')
    })


class EndpointSchema(ma.ModelSchema):
    class Meta:
        model = models.Endpoint
        fields = ('id', 'name', 'protocol', 'address', 'engines',
                  'tags', '_links')

    engines = ma.Nested('EngineSchema', many=True, exclude=('endpoints',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_endpoint', id='<id>'),
        'collection': ma.URLFor('show_endpoints')
    })


class UserSchema(ma.ModelSchema):
    class Meta:
        model = models.User
        fields = ('id', 'user', 'name', 'auth_key',
                  'auth_proto', 'priv_key', 'priv_proto',
                  'engines', 'tags', '_links')

    engines = ma.Nested('EngineSchema', many=True, exclude=('users',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_user', id='<id>'),
        'collection': ma.URLFor('show_users')
    })


class EngineSchema(ma.ModelSchema):
    class Meta:
        model = models.Engine
        fields = ('id', 'name', 'engine_id', 'users', 'endpoints',
                  'agents', 'tags', '_links')

    users = ma.Nested(UserSchema, many=True, exclude=('engines',))
    endpoints = ma.Nested(EndpointSchema, many=True, exclude=('engines',))
    agents = ma.Nested('AgentSchema', many=True, exclude=('engines',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_engine', id='<id>'),
        'collection': ma.URLFor('show_engines')
    })


class SelectorSchema(ma.ModelSchema):
    class Meta:
        model = models.Selector
        fields = ('id', 'comment', 'template', 'agents', 'order',
                  'tags', '_links')

    order = ma.Nested('AgentSelector')  # TODO: this does not work
    agents = ma.Nested('AgentSchema', many=True, exclude=('selectors',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_selector', id='<id>'),
        'collection': ma.URLFor('show_selectors')
    })


class AgentSchema(ma.ModelSchema):
    class Meta:
        model = models.Agent
        fields = ('id', 'name', 'engines', 'data_dir', 'selectors',
                  'labs', 'tags', '_links')

    engines = ma.Nested(EngineSchema, many=True, exclude=('agents',))
    selectors = ma.Nested(SelectorSchema, many=True, exclude=('agents',))
    labs = ma.Nested('LabSchema', many=True, exclude=('agents',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_agent', id='<id>'),
        'collection': ma.URLFor('show_agents')
    })


class RecordingSchema(ma.ModelSchema):
    class Meta:
        fields = ('path', 'size', 'type')


class LabSchema(ma.ModelSchema):
    class Meta:
        model = models.Lab
        fields = ('id', 'name', 'power', 'agents',
                  'tags', '_links')

    agents = ma.Nested(AgentSchema, many=True, exclude=('labs',))
    tags = ma.Nested('MinimalTagSchema', many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_lab', id='<id>'),
        'collection': ma.URLFor('show_labs')
    })


class TagSchema(MinimalTagSchema):
    class Meta:
        model = models.Tag
        fields = ('id', 'name', 'description', 'endpoints', 'users',
                  'engines', 'selectors', 'agents', 'labs', '_links')

    class EndpointSchema(EndpointSchema):
        class Meta:
            model = models.Endpoint
            fields = ('id', 'name', 'protocol', 'address', '_links')

    class UserSchema(UserSchema):
        class Meta:
            model = models.User
            fields = ('id', 'user', 'name', 'auth_key',
                      'auth_proto', 'priv_key', 'priv_proto',
                      '_links')

    class EngineSchema(EngineSchema):
        class Meta:
            model = models.Engine
            fields = ('id', 'name', 'engine_id', '_links')

    class SelectorSchema(SelectorSchema):
        class Meta:
            model = models.Selector
            fields = ('id', 'comment', 'template', 'order', '_links')

    class AgentSchema(AgentSchema):
        class Meta:
            model = models.Agent
            fields = ('id', 'name', 'data_dir', '_links')

    class LabSchema(LabSchema):
        class Meta:
            model = models.Lab
            fields = ('id', 'name', 'power', '_links')

    endpoints = ma.Nested(EndpointSchema, many=True, exclude=('tags',))
    users = ma.Nested(UserSchema, many=True, exclude=('tags',))
    engines = ma.Nested(EngineSchema, many=True, exclude=('tags',))
    selectors = ma.Nested(SelectorSchema, many=True, exclude=('tags',))
    agents = ma.Nested(AgentSchema, many=True, exclude=('tags',))
    labs = ma.Nested(LabSchema, many=True, exclude=('tags',))



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: ORM models
#
from snmpsim_control_plane.metrics import db


class Transport(db.Model):
    id = db.Column(db.Integer(), unique=True)
    transport_protocol = db.Column(db.String(8), nullable=False)
    endpoint = db.Column(db.String(64), nullable=False)
    peer = db.Column(db.String(64), nullable=False)

    packets = db.relationship(
        'Packet', cascade="all,delete", backref='transports', lazy=True)

    # Sqlalchemy's merge requires unique fields to be primary keys
    __table_args__ = (
        db.PrimaryKeyConstraint(
            'transport_protocol', 'endpoint', 'peer'
        ),
    )


class Packet(db.Model):
    total = db.Column(db.BigInteger)
    parse_failures = db.Column(db.BigInteger)
    auth_failures = db.Column(db.BigInteger)
    context_failures = db.Column(db.BigInteger)

    transport_id = db.Column(
        db.Integer, db.ForeignKey("transport.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint(
            'transport_id'
        ),
    )


class Agent(db.Model):
    id = db.Column(db.Integer(), unique=True)
    transport_id = db.Column(
        db.Integer, db.ForeignKey("transport.id"), nullable=False)
    engine = db.Column(db.String(64), nullable=False)
    security_model = db.Column(db.Integer(), nullable=False)
    security_level = db.Column(db.Integer(), nullable=False)
    context_engine = db.Column(db.String(64), nullable=False)
    context_name = db.Column(db.String(64), nullable=False)

    recordings = db.relationship(
        'Recording', cascade="all,delete", backref='agent', lazy=True)

    # Sqlalchemy's merge requires unique fields to be primary keys
    __table_args__ = (
        db.PrimaryKeyConstraint(
            'transport_id', 'engine', 'security_model', 'security_level',
            'context_engine', 'context_name'
        ),
    )


class Recording(db.Model):
    id = db.Column(db.Integer(), unique=True)
    path = db.Column(db.String(16), nullable=False)

    agent_id = db.Column(
        db.Integer, db.ForeignKey("agent.id"), nullable=False)

    pdus = db.relationship(
        'Pdu', cascade="all,delete", backref='recording', lazy=True)

    __table_args__ = (
        db.PrimaryKeyConstraint(
            'agent_id', 'path'
        ),
    )


class Pdu(db.Model):
    id = db.Column(db.Integer(), unique=True)
    name = db.Column(db.String(16), nullable=False)
    total = db.Column(db.BigInteger)

    recording_id = db.Column(
        db.Integer, db.ForeignKey("recording.id"), nullable=False)

    varbinds = db.relationship(
        'VarBind', cascade="all,delete", backref='pdu', lazy=True)
    variations = db.relationship(
        'Variation', cascade="all,delete", backref='pdu', lazy=True)

    __table_args__ = (
        db.PrimaryKeyConstraint(
            'recording_id', 'name'
        ),
    )


class VarBind(db.Model):
    total = db.Column(db.BigInteger)
    failures = db.Column(db.BigInteger)
    pdu_id = db.Column(
        db.Integer, db.ForeignKey("pdu.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint(
            'pdu_id'
        ),
    )


class Variation(db.Model):
    name = db.Column(db.String(64), nullable=False)
    total = db.Column(db.BigInteger)
    failures = db.Column(db.BigInteger)
    pdu_id = db.Column(
        db.Integer, db.ForeignKey("pdu.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint(
            'pdu_id', 'name'
        ),
    )


class Endpoint(db.Model):
    id = db.Column(db.Integer(), unique=True)
    protocol = db.Column(db.String(), nullable=False)
    address = db.Column(db.String(), nullable=False)
    process_id = db.Column(db.Integer(), db.ForeignKey('process.id'))

    __table_args__ = (
        db.PrimaryKeyConstraint('protocol', 'address', 'process_id'),
    )


class ConsolePage(db.Model):
    id = db.Column(db.Integer(), unique=True)
    text = db.Column(db.String(), nullable=False)
    timestamp = db.Column(db.DateTime(), nullable=False)
    process_id = db.Column(db.Integer(), db.ForeignKey('process.id'))

    __table_args__ = (
        db.PrimaryKeyConstraint('id', 'process_id'),
    )


class Process(db.Model):
    id = db.Column(db.Integer(), unique=True)
    path = db.Column(db.String(), nullable=False)
    runtime = db.Column(db.Integer())
    memory = db.Column(db.Integer())
    cpu = db.Column(db.Integer())
    files = db.Column(db.Integer())
    exits = db.Column(db.Integer())
    changes = db.Column(db.Integer())
    last_update = db.Column(db.DateTime())
    update_interval = db.Column(db.Integer())
    supervisor_id = db.Column(db.Integer(), db.ForeignKey('supervisor.id'))

    endpoints = db.relationship(
        'Endpoint', backref='process', lazy=True)
    console_pages = db.relationship(
        'ConsolePage', backref='process', lazy=True)

    __table_args__ = (
        db.PrimaryKeyConstraint('supervisor_id', 'path'),
    )


class Supervisor(db.Model):
    id = db.Column(db.Integer(), unique=True)
    hostname = db.Column(db.String(), nullable=False)
    watch_dir = db.Column(db.String(), nullable=False)
    started = db.Column(db.DateTime())

    processes = db.relationship(
        'Process', backref='supervisor', lazy=True)

    __table_args__ = (
        db.PrimaryKeyConstraint('hostname'),
    )



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: REST API management server
#
import argparse
import os
import sys

from snmpsim_control_plane.management import app
from snmpsim_control_plane.management import db
from snmpsim_control_plane.management import views  # noqa

DESCRIPTION = """\
SNMP Simulation Control Plane REST API server.
 
Maintains SNMP simulator configuration in a DB manageable via REST API.
Generates SNMP simulator invocation script with necessary parameter
realizing the most rest known configuration.

Can be run as a WSGI application.
"""


def parse_args():
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        '--recreate-db',
        action='store_true',
        help='DANGER! Running with this flag wipes up REST API server DB! '
             'This switch makes sense only when running this tool for the '
             'first time.')

    parser.add_argument(
        '--config', type=str,
        help='Config file path. Can also be set via environment variable '
             'SNMPSIM_MGMT_CONFIG.')

    parser.add_argument(
        '--interface', type=str,
        help='IP address of the local interface for REST API'
             'server to listen on. Can also be set via config variable '
             'SNMPSIM_MGMT_LISTEN_IP. Default is all local interfaces.')

    parser.add_argument(
        '--port', type=int,
        help='TCP port to bind REST API server to.  Can also be '
             'set via config variable SNMPSIM_MGMT_LISTEN_PORT. '
             'Default is 5000.')

    parser.add_argument(
        '--data-root', type=str,
        help='Path to a SNMP simulation data root directory. SNMP command '
             'responder will be configured with its data files being under '
             'this directory. Can also be set via config variable '
             'SNMPSIM_MGMT_DATAROOT.')

    parser.add_argument(
        '--template', type=str,
        help='Path to Jinja2 template to use for rendering SNMP simulator command '
             'responder invocation command. Can also be set via config variable '
             'SNMPSIM_MGMT_TEMPLATE. By default, built-in template is used.')

    parser.add_argument(
        '--destination', type=str,
        help='Path to a directory where REST API server will place rendered '
             'SNMP simulator command responder invocation scripts. Can also '
             'be set via config variable SNMPSIM_MGMT_DESTINATION.')

    return parser.parse_args()


def main():

    args = parse_args()

    if args.config:
        os.environ['SNMPSIM_MGMT_CONFIG'] = args.config

    config_file = os.environ.get('SNMPSIM_MGMT_CONFIG')
    if config_file:
        app.config.from_pyfile(config_file)

    if args.interface:
        app.config['SNMPSIM_MGMT_LISTEN_IP'] = args.interface

    if args.port:
        app.config['SNMPSIM_MGMT_LISTEN_PORT'] = args.port

    if args.data_root:
        app.config['SNMPSIM_MGMT_DATAROOT'] = args.data_root

    if args.template:
        app.config['SNMPSIM_MGMT_TEMPLATE'] = args.template

    if args.destination:
        app.config['SNMPSIM_MGMT_DESTINATION'] = args.destination

    if args.recreate_db:
        db.drop_all()
        db.create_all()
        return 0

    app.run(host=app.config.get('SNMPSIM_MGMT_LISTEN_IP'),
            port=app.config.get('SNMPSIM_MGMT_LISTEN_PORT'))

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: REST API document model
#
import re

from sqlalchemy.orm import validates

from snmpsim_control_plane.management import db


class Tag(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(32), nullable=False, unique=True)
    description = db.Column(db.String(), nullable=True)

    endpoints = db.relationship(
        'Endpoint', secondary='tag_endpoint', back_populates='tags', lazy=True)
    users = db.relationship(
        'User', secondary='tag_user', back_populates='tags', lazy=True)
    engines = db.relationship(
        'Engine', secondary='tag_engine', back_populates='tags', lazy=True)
    selectors = db.relationship(
        'Selector', secondary='tag_selector', back_populates='tags', lazy=True)
    agents = db.relationship(
        'Agent', secondary='tag_agent', back_populates='tags', lazy=True)
    labs = db.relationship(
        'Lab', secondary='tag_lab', back_populates='tags', lazy=True)


class Endpoint(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(), nullable=True)
    protocol = db.Column(db.Enum('udpv4', 'udpv6'), nullable=False)
    address = db.Column(db.String(64), unique=True, nullable=False)
    engines = db.relationship(
        'Engine', secondary='engine_endpoint',
        back_populates='endpoints', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_endpoint', back_populates='endpoints', lazy=True)

    @validates('address')
    def validate_address(self, key, address):
        # TODO: also match `protocol`
        ipv4 = re.match(r'[0-9]+(?:\.[0-9]+){3}(:[0-9]+)?$', address)
        if ipv4:
            return address

        ipv6 = re.match(r'\[([a-f0-9:]+:+)+[a-f0-9]+\](:[0-9]+)?$', address)
        if ipv6:
            return address

        raise Exception('Malformed IP address %s' % address)


class TagEndpoint(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    endpoint_id = db.Column(
        db.Integer, db.ForeignKey("endpoint.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'endpoint_id'),
    )


class User(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    user = db.Column(db.String(32), unique=True)
    name = db.Column(db.String(60))
    auth_key = db.Column(db.String(), nullable=True)
    auth_proto = db.Column(
        db.Enum("MD5", "SHA", "SHA224", "SHA256", "SHA384", "SHA512",
                "NONE"), default='NONE')
    priv_key = db.Column(db.String(), nullable=True)
    priv_proto = db.Column(
        db.Enum("DES", "3DES", "AES", "AES128", "AES192", "AES192BLMT",
                "AES256", "AES256BLMT", "NONE"), default='NONE')
    engines = db.relationship(
        'Engine', secondary='engine_user', back_populates='users', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_user', back_populates='users', lazy=True)

    @validates('auth_proto')
    def uppercase_auth_proto(self, key, proto):
        return proto.upper()

    @validates('priv_proto')
    def uppercase_priv_proto(self, key, proto):
        return proto.upper()

    __table_args__ = (
        db.CheckConstraint(
            "auth_key IS NULL AND auth_proto='NONE' OR "
            "auth_key IS NOT NULL AND length(auth_key) > 7 "
            "AND auth_proto!='NONE'",
            name='auth_key_check'),
        db.CheckConstraint(
            "priv_key IS NULL AND priv_proto='NONE' OR "
            "priv_key IS NOT NULL AND length(priv_key) > 7 "
            "AND priv_proto!='NONE'",
            name='priv_key_check'),
        db.CheckConstraint(
            "priv_proto='NONE' OR "
            "priv_proto!='NONE' AND auth_proto!='NONE'",
            name='auth_priv_check'),
    )


class TagUser(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    user_id = db.Column(
        db.Integer, db.ForeignKey("user.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'user_id'),
    )


class Engine(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(64), nullable=True)
    engine_id = db.Column(db.String(32), default='auto')
    users = db.relationship(
        'User', secondary='engine_user',
        back_populates='engines', lazy=True)
    agents = db.relationship(
        'Agent', secondary='agent_engine',
        back_populates='engines', lazy=True)
    endpoints = db.relationship(
        'Endpoint', secondary='engine_endpoint',
        back_populates='engines', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_engine', back_populates='engines', lazy=True)


class EngineUser(db.Model):
    engine_id = db.Column(
        db.Integer, db.ForeignKey("engine.id"), nullable=False)
    user_id = db.Column(
        db.Integer, db.ForeignKey("user.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('engine_id', 'user_id'),
    )


class EngineEndpoint(db.Model):
    engine_id = db.Column(
        db.Integer, db.ForeignKey("engine.id"), nullable=False)
    endpoint_id = db.Column(
        db.Integer, db.ForeignKey("endpoint.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('engine_id', 'endpoint_id'),
    )


class TagEngine(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    engine_id = db.Column(
        db.Integer, db.ForeignKey("engine.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'engine_id'),
    )


class Selector(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    comment = db.Column(db.String())
    template = db.Column(db.String())
    agents = db.relationship(
        'Agent', backref='selector', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_selector', back_populates='selectors', lazy=True)


class TagSelector(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    selector_id = db.Column(
        db.Integer, db.ForeignKey("selector.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'selector_id'),
    )


class Agent(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(), nullable=True)
    data_dir = db.Column(db.String(), default='.')
    engine_id = db.Column(db.Integer, db.ForeignKey('engine.id'))
    selector_id = db.Column(db.Integer, db.ForeignKey('selector.id'))
    engines = db.relationship(
        'Engine', secondary='agent_engine', back_populates='agents', lazy=True)
    selectors = db.relationship(
        'Selector', secondary='agent_selector', back_populates='agents', lazy=True)
    labs = db.relationship(
        'Lab', secondary='lab_agent', back_populates='agents', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_agent', back_populates='agents', lazy=True)


class AgentEngine(db.Model):
    engine_id = db.Column(
        db.Integer, db.ForeignKey("engine.id"), nullable=False)
    agent_id = db.Column(
        db.Integer, db.ForeignKey("agent.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('engine_id', 'agent_id'),
    )


class AgentSelector(db.Model):
    order = db.Column(db.Integer())
    selector_id = db.Column(
        db.Integer, db.ForeignKey("selector.id"), nullable=False)
    agent_id = db.Column(
        db.Integer, db.ForeignKey("agent.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('selector_id', 'agent_id'),
    )


class TagAgent(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    agent_id = db.Column(
        db.Integer, db.ForeignKey("agent.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'agent_id'),
    )


class Lab(db.Model):
    id = db.Column(db.Integer(), primary_key=True)
    name = db.Column(db.String(), nullable=True)
    power = db.Column(db.Enum('on', 'off'), default='off')
    agents = db.relationship(
        'Agent', secondary='lab_agent', back_populates='labs', lazy=True)
    tags = db.relationship(
        'Tag', secondary='tag_lab', back_populates='labs', lazy=True)


class LabAgent(db.Model):
    lab_id = db.Column(
        db.Integer, db.ForeignKey("lab.id"), nullable=False)
    agent_id = db.Column(
        db.Integer, db.ForeignKey("agent.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('lab_id', 'agent_id'),
    )


class TagLab(db.Model):
    tag_id = db.Column(
        db.Integer, db.ForeignKey("tag.id"), nullable=False)
    lab_id = db.Column(
        db.Integer, db.ForeignKey("lab.id"), nullable=False)

    __table_args__ = (
        db.PrimaryKeyConstraint('tag_id', 'lab_id'),
    )



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: Do not report supervisor metrics
#
from snmpsim_control_plane.supervisor.reporting.formats import base


class NullReporter(base.BaseReporter):
    """No-op activity metrics reporter.
    """



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: supervisor metrics importer
#
import datetime
import time

from snmpsim_control_plane.metrics import db
from snmpsim_control_plane.metrics import models
from snmpsim_control_plane.metrics.utils import autoincrement

MAX_CONSOLE_PAGE_AGE = 86400  # one day


def import_metrics(jsondoc):
    """Update metrics DB from `dict` data structure

    The input data structure is expected to be the one produced by SNMP
    simulator's command responder `fulljson` reporting module.



    .. code-block:: python

    {
        'format': 'jsondoc',
        'version': 1,
        'host': '{hostname}',
        'watch_dir': {dir},
        'started': '{timestamp}',
        'first_update': '{timestamp}',
        'last_update': '{timestamp}',
        'executables': [
            {
                'executable': '{path}',
                'runtime': '{seconds}',
                'memory': {MB},
                'cpu': {ms},
                'files': 0,
                'exits': 0,
                'changes': 0,
                'endpoints': {
                    'udpv4': [
                        '127.0.0.1:161'
                    ],
                    'udpv6': [
                        '::1:161'
                    ]
                ],
                'console': [
                    {
                        'timestamp': 0,
                        'text': '{text}'
                    }
                ]
            }
        ]
    }
    """
    old_times = int(time.time() - MAX_CONSOLE_PAGE_AGE)

    timestamp = datetime.datetime.utcfromtimestamp(
        jsondoc['started'])

    supervisor_model = models.Supervisor(
        hostname=jsondoc['host'],
        watch_dir=jsondoc['watch_dir'],
        started=timestamp
    )

    supervisor_model = db.session.merge(supervisor_model)

    autoincrement(supervisor_model, models.Supervisor)

    for executable in jsondoc['executables']:

        process_model = models.Process(
            path=executable['executable'],
            supervisor_id=supervisor_model.id
        )

        process_model = db.session.merge(process_model)

        autoincrement(process_model, models.Process)

        process_model.runtime = process_model.runtime or 0
        process_model.runtime += executable['runtime']

        process_model.memory = executable['memory']

        process_model.cpu = process_model.cpu or 0
        process_model.cpu += executable['cpu']

        process_model.files = executable['files']

        process_model.exits = process_model.exits or 0
        process_model.exits += executable['exits']

        process_model.changes = process_model.changes or 0
        process_model.changes += executable['changes']

        process_model.update_interval = (
                jsondoc['last_update'] - jsondoc['first_update'])

        timestamp = datetime.datetime.utcfromtimestamp(
            jsondoc['last_update'])

        process_model.last_update = timestamp

        query = (
            models.Endpoint
            .query
            .filter_by(process_id=process_model.id))

        existing_endpoints = set(
            (x.protocol, x.address) for x in query.all())

        reported_endpoints = set()

        for protocol, addresses in executable['endpoints'].items():
            for address in addresses:
                reported_endpoints.add((protocol, address))

        new_endpoints = reported_endpoints.difference(existing_endpoints)

        for protocol, address in new_endpoints:
            endpoint_model = models.Endpoint(
                protocol=protocol,
                address=address,
                process_id=process_model.id
            )

            autoincrement(endpoint_model, models.Endpoint)
            db.session.add(endpoint_model)

        removed_endpoints = existing_endpoints.difference(reported_endpoints)

        for protocol, address in removed_endpoints:
            query = (
                db.session
                .query(models.Endpoint)
                .filter_by(protocol=protocol)
                .filter_by(address=address)
                .filter_by(process_id=process_model.id))

            query.delete()

        query = (
            db.session
            .query(models.ConsolePage)
            .filter(models.ConsolePage.process_id == process_model.id)
            .filter(models.ConsolePage.timestamp < old_times)
        )

        query.delete()

        for console_page in executable['console']:

            timestamp = datetime.datetime.utcfromtimestamp(
                console_page['timestamp'])

            console_page_model = models.ConsolePage(
                timestamp=timestamp,
                text=console_page['text'],
                process_id=process_model.id
            )

            autoincrement(console_page_model, models.ConsolePage)

            db.session.add(console_page_model)

        db.session.commit()



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: Report supervisor metrics in JSON form
#
import json
import os
import socket
import tempfile
import uuid

from snmpsim_control_plane import error
from snmpsim_control_plane import log
from snmpsim_control_plane.supervisor import lifecycle
from snmpsim_control_plane.supervisor.reporting.formats import base


class JsonDocReporter(base.BaseReporter):
    """Dump metrics as a JSON document,

    Metrics are arranged as a data structure like this:

    .. code-block:: python

    {
        'format': 'jsondoc',
        'version': 1,
        'host': '{hostname}',
        'producer': <UUID>,
        'watch_dir': '{dir}',
        'uptime': 0,
        'first_update': '{timestamp}',
        'last_update': '{timestamp}',
        'executables': [
            {
                'executable': '{path}',
                'runtime': '{seconds}',
                'memory': 0,
                'cpu': 0,
                'files': 0,
                'exits': 0,
                'changes': 0,
                'endpoints': {
                    'udpv4': [
                        '127.0.0.1:161'
                    ],
                    'udpv6': [
                        '::1:161'
                    ]
                ],
                'console': [
                    {
                        'timestamp': 0,
                        'text': '{text}'
                    }
                ]
            }
        ]
    }

    """
    REPORTING_FORMAT = 'jsondoc'
    REPORTING_VERSION = 1
    PRODUCER_HOST = socket.gethostname()
    PRODUCER_UUID = str(uuid.uuid1())

    def __init__(self, *args, **kwargs):
        super(JsonDocReporter, self).__init__(*args, **kwargs)

        if not args:
            raise error.ControlPlaneError(
                'Missing %s parameter(s). Expected: '
                '<method>:<reports-dir>[:dumping-period]' % self.__class__.__name__)

        self._reports_dir = os.path.join(args[0], self.REPORTING_FORMAT)

        if len(args) > 1:
            try:
                self.REPORTING_PERIOD = int(args[1])

            except Exception as exc:
                raise error.ControlPlaneError(
                    'Malformed reports dumping period: %s' % args[1])

        try:
            if not os.path.exists(self._reports_dir):
                os.makedirs(self._reports_dir)

        except OSError as exc:
            raise error.ControlPlaneError(
                'Failed to create reports directory %s: '
                '%s' % (self._reports_dir, exc))

        log.debug(
            'Initialized %s metrics reporter for instance %s, metrics '
            'directory %s' % (
                self.__class__.__name__, self.PRODUCER_UUID, self._reports_dir))

    @staticmethod
    def _json_serializer(obj):
        if isinstance(obj, lifecycle.ConsoleLog):
            return [
                {'page': page,
                 'text': obj.text(page),
                 'timestamp': obj.timestamp(page)}
                for page in range(obj.first_page, obj.last_page + 1)
            ]

        return obj

    def dump_metrics(self, metrics, watch_dir=None,
                     started=None, begin=None, end=None):
        """Dump metrics into a JSON file.
        """
        json_metrics = self._format_metrics(metrics)

        json_metrics['format'] = self.REPORTING_FORMAT
        json_metrics['version'] = self.REPORTING_VERSION
        json_metrics['host'] = self.PRODUCER_HOST
        json_metrics['producer'] = self.PRODUCER_UUID
        json_metrics['watch_dir'] = watch_dir
        json_metrics['started'] = started
        json_metrics['first_update'] = int(begin)
        json_metrics['last_update'] = int(end)

        dump_path = os.path.join(self._reports_dir, '%s.json' % end)

        log.debug('Dumping JSON metrics to %s' % dump_path)

        try:
            json_doc = json.dumps(
                json_metrics, indent=2, default=self._json_serializer)

            with tempfile.NamedTemporaryFile(delete=False) as fl:
                fl.write(json_doc.encode('utf-8'))

            os.rename(fl.name, dump_path)

        except Exception as exc:
            log.error(
                'Failure while dumping metrics into '
                '%s: %s' % (dump_path, exc))

    @staticmethod
    def _format_metrics(metrics):
        """Reformat generic metrics into a specific layout
        """
        return {'executables': metrics}



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: snmpsim metrics helpers
#
from sqlalchemy import func

from snmpsim_control_plane.metrics import db


def autoincrement(obj, model):
    """Add unique ID to model.

    Sqlalchemy's merge requires unique fields being primary keys. On top of
    that, autoincrement does not always work with Sqlalchemy. Thus this
    hack to generate unique row ID. %-(
    """
    if obj.id is None:
        max_id = db.session.query(func.max(model.id)).first()
        max_id = max_id[0]
        max_id = max_id + 1 if max_id else 1
        obj.id = max_id

        db.session.commit()




#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator Control Plane: metrics importer manager
#
from snmpsim_control_plane.metrics import db

from snmpsim_control_plane import log
from snmpsim_control_plane.metrics.importers import snmpagent
from snmpsim_control_plane.metrics.importers import process

KNOWN_IMPORTERS = {
    'fulljson': snmpagent.import_metrics,
    'jsondoc': process.import_metrics,
}


def import_metrics(jsondoc):
    """Update metrics DB from `dict` data structure

    The input data structure is expected to be the one produced by SNMP
    simulator's command responder `fulljson` reporting module.
    """
    flavor = jsondoc.get('format')
    importer = KNOWN_IMPORTERS.get(flavor)
    if not importer:
        log.error('Unknown metric flavor %s, '
                  'ignoring' % flavor or '<unspecified>')
        return

    try:
        importer(jsondoc)

    except Exception as exc:
        log.error('Metric importer %s failed: %s' % (flavor, exc))
        log.error('JSON document causing failure is: %s' % jsondoc)
        db.session.rollback()



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP Agent Simulator: REST API metrics server
#


class DefaultConfig(object):
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory'
    SQLALCHEMY_ECHO = False

    DEBUG = False

    SNMPSIM_METRICS_LISTEN_IP = '127.0.0.1'
    SNMPSIM_METRICS_LISTEN_PORT = 5001
    SNMPSIM_METRICS_SSL_CERT = None
    SNMPSIM_METRICS_SSL_KEY = None



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator management: data simulation recording management
#
import os

from snmpsim_control_plane import log
from snmpsim_control_plane import error


KNOWN_EXTENSIONS = {
    '.snmprec': 'snmprec',
    '.sapwalk': 'sapwalk',
    '.snmpwalk': 'snmpwalk'
}


def _traverse_dir(directory):
    files = []
    entries = os.listdir(directory)

    for entry in entries:
        dir_or_file = os.path.join(directory, entry)
        if os.path.isdir(dir_or_file):
            files.extend(_traverse_dir(dir_or_file))

        else:
            files.append(dir_or_file)

    return files


def get_recording_type(filename):
    for ext in KNOWN_EXTENSIONS:
        if filename.endswith(ext):
            return KNOWN_EXTENSIONS[ext]


def list_recordings(data_dir):

    try:
        files = _traverse_dir(data_dir)

    except Exception as exc:
        raise error.ControlPlaneError(
            'Directory %s traversal failure: %s' % (data_dir, exc))

    recordings = []

    for filename in files:
        stat = os.stat(filename, follow_symlinks=True)

        recording_type = get_recording_type(filename)
        if not recording_type:
            continue

        recording = {
            'path': filename[len(data_dir) + 1:],
            'size': stat.st_size,
            'type': recording_type
        }

        recordings.append(recording)

    return recordings


def get_recording(data_dir, filename, exists=False,
                  not_exists=False, ensure_path=False):
    filename = os.path.abspath(os.path.join(data_dir, filename))

    if not filename.startswith(data_dir):
        log.error('Requested recording %s is outside of data root' % filename)
        raise error.ControlPlaneError('No such recording')

    if exists and not os.path.exists(filename):
        log.error('Requested recording %s does not exist' % filename)
        raise error.ControlPlaneError('No such recording')

    if not_exists and os.path.exists(filename):
        log.error('Requested recording %s unexpectedly exists' % filename)
        raise error.ControlPlaneError('No such recording')

    if ensure_path:
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            try:
                os.makedirs(directory)

            except OSError as exc:
                log.error('Failed to create %s: exc' % directory)
                raise error.ControlPlaneError('No such recording')

        elif not os.path.isdir(directory):
            raise error.ControlPlaneError('No such recording')

    return os.path.dirname(filename), os.path.basename(filename)



#
# This file is part of SNMP Simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
import sys

from snmpsim_control_plane import error

if sys.platform[:3] == 'win':

    def daemonize(pidfile):
        raise error.ControlPlaneError(
            'Windows is not inhabited with daemons!')

else:

    import os
    import atexit
    import signal
    import tempfile


    def daemonize(pidfile):
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                os._exit(0)

        except OSError as exc:
            raise error.ControlPlaneError(
                'ERROR: fork #1 failed: %s' % exc)

        # decouple from parent environment
        try:
            os.chdir('/')
            os.setsid()

        except OSError:
            pass

        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                os._exit(0)

        except OSError as exc:
            raise error.ControlPlaneError(
                'ERROR: fork #2 failed: %s' % exc)

        def signal_cb(s, f):
            raise KeyboardInterrupt

        for s in (signal.SIGTERM, signal.SIGINT, signal.SIGHUP,
                  signal.SIGQUIT):
            signal.signal(s, signal_cb)

        # write pidfile
        def atexit_cb():
            try:
                if pidfile:
                    os.remove(pidfile)

            except OSError:
                pass

        atexit.register(atexit_cb)

        try:
            if pidfile:
                fd, nm = tempfile.mkstemp(dir=os.path.dirname(pidfile))
                os.write(fd, ('%d\n' % os.getpid()).encode('utf-8'))
                os.close(fd)
                os.rename(nm, pidfile)

        except Exception as exc:
            raise error.ControlPlaneError(
                'Failed to create PID file %s: %s' % (pidfile, exc))

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())



#
# This file is part of snmpsim-control-plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim-control-plane/license.html
#
"""SNMP Simulator Control Plane

REST API driven management and monitoring supervisor to
remotely operate SNMP simulator.
"""
import os
import sys

import setuptools

classifiers = """\
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Developers
Intended Audience :: Information Technology
Intended Audience :: System Administrators
Intended Audience :: Telecommunications Industry
License :: OSI Approved :: BSD License
Natural Language :: English
Operating System :: OS Independent
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.2
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Communications
Topic :: System :: Monitoring
Topic :: System :: Networking :: Monitoring
"""

if sys.version_info[:2] < (2, 7):
    print("ERROR: this package requires Python 2.7 or later!")
    sys.exit(1)

with open(os.path.join('snmpsim_control_plane', '__init__.py')) as fl:
    version = fl.read().split('\'')[1]

# make sure to have fresh `setuptools` for this to work
with open('requirements.txt') as fl:
    requirements = fl.readlines()

doclines = [x.strip() for x in (__doc__ or '').split('\n') if x]

params = {
    'name': 'snmpsim-control-plane',
    'version': version,
    'description': doclines[0],
    'long_description': ' '.join(doclines[1:]),
    'maintainer': 'Ilya Etingof <etingof@gmail.com>',
    'author': 'Ilya Etingof',
    'author_email': 'etingof@gmail.com',
    'url': 'http://snmplabs.com/snmpsim-control-plane',
    'license': 'BSD',
    'platforms': ['any'],
    'classifiers': [x for x in classifiers.split('\n') if x],
    'packages': setuptools.find_packages(),
    'entry_points': {
        'console_scripts': [
            'snmpsim-mgmt-restapi = snmpsim_control_plane.commands.management:main',
            'snmpsim-mgmt-supervisor = snmpsim_control_plane.commands.supervisor:main',
            'snmpsim-metrics-importer = snmpsim_control_plane.commands.importer:main',
            'snmpsim-metrics-restapi = snmpsim_control_plane.commands.metrics:main',
        ]
    },
    'install_requires': requirements,
    'zip_safe': False,  # Flask
    'include_package_data': True
}

setuptools.setup(**params)



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: REST API views
#
import flask
from werkzeug import exceptions
from sqlalchemy import func

from snmpsim_control_plane.metrics import app
from snmpsim_control_plane.metrics import models
from snmpsim_control_plane.metrics import schemas

PREFIX = '/snmpsim/metrics/v1'


@app.errorhandler(exceptions.HTTPException)
def flask_exception_handler(exc):
    app.logger.error(exc)
    err = {
        'status': exc.code,
        'message': exc.description
    }
    response = flask.jsonify(err)
    response.status_code = exc.code
    return response


@app.errorhandler(Exception)
def all_exception_handler(exc):
    app.logger.error(exc)
    err = {
        'status': 400,
        'message': getattr(exc, 'message', str(exc))
    }
    response = flask.jsonify(err)
    response.status_code = 400
    return response


PACKETS_QS_COLUMN_MAP = {
    'protocol': models.Transport.transport_protocol,
    'local_address': models.Transport.endpoint,
    'peer_address': models.Transport.peer,
}

MESSAGES_QS_COLUMN_MAP = {
    'engine_id': models.Agent.engine,
    'security_model': models.Agent.security_model,
    'security_level': models.Agent.security_level,
    'context_engine_id': models.Agent.context_engine,
    'context_name': models.Agent.context_name,
    'pdu_type': models.Pdu.name,
    'recording': models.Recording.path,
}

QS_COLUMN_MAP = PACKETS_QS_COLUMN_MAP.copy()
QS_COLUMN_MAP.update(MESSAGES_QS_COLUMN_MAP)


def filter_by(query, *fields):
    search_columns = flask.request.args

    unknown_columns = set(search_columns).difference(fields)
    if unknown_columns:
        raise exceptions.NotFound(
            'Search term(s) %s not supported' % ', '.join(unknown_columns))

    for field in fields:
        args = search_columns.getlist(field)
        if args:
            query = query.filter(
                QS_COLUMN_MAP[field].in_(args))

    return query


@app.route(PREFIX + '/')
def show_root():
    return {
        'activity': flask.url_for('show_activity')
    }


@app.route(PREFIX + '/activity')
def show_activity():
    return {
        'packets': flask.url_for('show_packets'),
        'messages': flask.url_for('show_messages')
    }


def _make_filter_hyperlinks(target):
    return {
        flt: {
            'self': {
                '_links': flask.url_for('show_filter', target=target, flt=flt)
            }
        }
        for flt in (PACKETS_QS_COLUMN_MAP
                    if target == 'packets' else QS_COLUMN_MAP)
    }


@app.route(PREFIX + '/activity/packets/filters')
def show_packets_filters():
    return _make_filter_hyperlinks(target='packets')


@app.route(PREFIX + '/activity/messages/filters')
def show_messages_filters():
    return _make_filter_hyperlinks(target='messages')


@app.route(PREFIX + '/activity/<target>/filters/<flt>')
def show_filter(target, flt):
    try:
        column = (PACKETS_QS_COLUMN_MAP
                  if target == 'packets' else QS_COLUMN_MAP)[flt]

    except KeyError:
        raise exceptions.NotFound('No such filter')

    metrics = (
        models.Transport
        .query
        .with_entities(column)
        .group_by(column)
        .all())

    return flask.jsonify([mtr[0] for mtr in metrics])


def _show_packets_or_messages(show_messages=False):
    transport_query = (
        models.Transport
        .query
        .with_entities(
            func.sum(models.Packet.total).label("total"),
            func.sum(models.Packet.parse_failures).label("parse_failures"),
            func.sum(models.Packet.auth_failures).label("auth_failures"),
            func.sum(models.Packet.context_failures).label("context_failures")))

    # We have to build JSON response by hand because here it's a mix of
    # ORM models and custom dicts. Marshmallow does not seem to be well-suited
    # for handling that.
    metrics = {}

    if show_messages:
        agent_query = (
            transport_query
            .join(models.Transport)
            .join(models.Agent)
            .join(models.Recording)
            .join(models.Pdu))

        agent_query = filter_by(agent_query, *QS_COLUMN_MAP)

        var_binds_query = (
            agent_query
            .join(models.VarBind)
            .with_entities(
                func.sum(models.Pdu.total).label("pdus"),
                func.sum(models.VarBind.total).label("var_binds"),
                func.sum(models.VarBind.failures).label("failures")))

        variations_query = (
            agent_query
            .join(models.Variation)
            .with_entities(
                models.Variation.name,
                func.sum(models.Variation.total).label("total"),
                func.sum(models.Variation.failures).label("failures"))
            .group_by(models.Variation.name))

        messages = var_binds_query.first()
        schema = schemas.MessagesSchema()
        messages = schema.dump(messages).data

        variations = variations_query.all()
        schema = schemas.VariationsSchema(many=True)
        variations = schema.dump(variations).data

        _self = flask.url_for(
            'show_messages', **dict(
                (field, flask.request.args.getlist(field))
                for field in QS_COLUMN_MAP))

        links = {
            'self': _self
        }

        filters = {
            '_links': {
                'self': flask.url_for('show_messages_filters')
            }
        }

        metrics.update(
            variations=variations, _links=links, filters=filters, **messages)

    else:
        transport_query = filter_by(transport_query, *PACKETS_QS_COLUMN_MAP)

        packets = transport_query.first()
        schema = schemas.PacketsSchema()
        packets = schema.dump(packets).data

        _self = flask.url_for(
            'show_packets', **dict(
                (field, flask.request.args.getlist(field))
                for field in PACKETS_QS_COLUMN_MAP))

        links = {
            'self': _self
        }

        filters = {
            '_links': {
                'self': flask.url_for('show_packets_filters')
            }
        }

        metrics.update(_links=links, filters=filters, **packets)

    return metrics


@app.route(PREFIX + '/activity/packets')
def show_packets():
    return _show_packets_or_messages(show_messages=False)


@app.route(PREFIX + '/activity/messages')
def show_messages():
    return _show_packets_or_messages(show_messages=True)


@app.route(PREFIX + '/processes')
@app.route(PREFIX + '/processes/<id>')
@app.route(PREFIX + '/supervisors/<supervisor_id>/processes')
def show_processes(id=None, supervisor_id=None):
    process_query = (
        models.Process
        .query
        .outerjoin(models.Endpoint)
        .outerjoin(models.ConsolePage))

    if supervisor_id is not None:
        process_query = (
            process_query
            .filter(models.Process.supervisor_id == supervisor_id))

    if id is None:
        processes = process_query.all()

    else:
        process_query = (
            process_query
            .filter(models.Process.id == id))

        processes = process_query.first()

        if not processes:
            raise exceptions.NotFound('Process not found')

    schema = schemas.ProcessSchema(many=id is None)
    return schema.jsonify(processes), 200


@app.route(PREFIX + '/supervisors')
@app.route(PREFIX + '/supervisors/<id>')
def show_supervisors(id=None):
    supervisor_query = (
        models.Supervisor
        .query
        .outerjoin(models.Process))

    if id is None:
        supervisors = supervisor_query.all()

    else:
        supervisor_query = (
            supervisor_query
            .filter(models.Supervisor.id == id))

        supervisors = supervisor_query.first()

        if not supervisors:
            raise exceptions.NotFound('Supervisor not found')

    schema = schemas.SupervisorSchema(many=id is None)
    return schema.jsonify(supervisors), 200


@app.route(PREFIX + '/endpoints')
@app.route(PREFIX + '/endpoints/<endpoint_id>')
@app.route(PREFIX + '/processes/<id>/endpoints')
@app.route(PREFIX + '/processes/<id>/endpoints/<endpoint_id>')
def show_endpoints(id=None, endpoint_id=None):
    endpoint_query = (
        models.Endpoint
        .query
        .join(models.Process))

    if id is not None:
        endpoint_query = (
            endpoint_query
            .filter(models.Process.id == id))

    if endpoint_id is None:
        endpoints = endpoint_query.all()

    else:
        endpoint_query = (
            endpoint_query
            .filter(models.Endpoint.id == id))

        endpoints = endpoint_query.first()

        if not endpoints:
            raise exceptions.NotFound('Endpoint not found')

    schema = schemas.EndpointSchema(many=endpoint_id is None)
    return schema.jsonify(endpoints), 200


@app.route(PREFIX + '/consoles/<id>')
@app.route(PREFIX + '/consoles/<id>/page/<page_id>')
@app.route(PREFIX + '/processes/<id>/console')
@app.route(PREFIX + '/processes/<id>/console/<page_id>')
def show_console(id, page_id=None):
    console_query = (
        models.ConsolePage
        .query
        .join(models.Process)
        .filter(models.Process.id == id)
        .order_by(models.ConsolePage.timestamp.asc()))

    if page_id is None:
        pages = console_query.all()

    else:
        console_query = (
            console_query
            .filter(models.ConsolePage.id == page_id))

        pages = console_query.first()

        if not pages:
            raise exceptions.NotFound('Console page not found')

    schema = schemas.ConsoleSchema(many=page_id is None)
    return schema.jsonify(pages), 200



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: snmpsim metrics importer
#
from snmpsim_control_plane.metrics import db
from snmpsim_control_plane.metrics import models
from snmpsim_control_plane.metrics.utils import autoincrement


def import_metrics(fulljson):
    """Update metrics DB from `dict` data structure

    The input data structure is expected to be the one produced by SNMP
    simulator's command responder `fulljson` reporting module.


    .. code-block:: python

    {
        'format': 'fulljson',
        'version': 1,
        'producer': <UUID>,
        'first_update': '{timestamp}',
        'last_update': '{timestamp}',
        '{transport_protocol}': {
            '{transport_endpoint}': {  # local address
                'transport_domain': '{transport_domain}',  # endpoint ID
                '{transport_address}', { # peer address
                    'packets': 0,
                    'parse_failures': 0,  # n/a
                    'auth_failures': 0,  # n/a
                    'context_failures': 0, # n/a
                    '{snmp_engine}': {
                        '{security_model}': {
                            '{security_level}': {
                                '{security_name}': {
                                    '{context_engine_id}': {
                                        '{context_name}': {
                                            '{pdu_type}': {
                                                '{data_file}': {
                                                    'pdus': 0,
                                                    'varbinds': 0,
                                                    'failures': 0,
                                                    '{variation_module}': {
                                                        'calls': 0,
                                                        'failures': 0
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    """

    for transport_protocol, transport_endpoints in fulljson.items():

        if not isinstance(transport_endpoints, dict):
            continue

        for transport_endpoint, peers in transport_endpoints.items():

            if not isinstance(peers, dict):
                continue

            for peer_adddress, engines in peers.items():

                if not isinstance(engines, dict):
                    continue

                transport_model = models.Transport(
                    transport_protocol=transport_protocol,
                    endpoint=transport_endpoint,
                    peer=peer_adddress,
                )

                transport_model = db.session.merge(transport_model)

                autoincrement(transport_model, models.Transport)

                packet_model = models.Packet(
                    transport_id=transport_model.id)

                packet_model = db.session.merge(packet_model)

                packet_model.total = packet_model.total or 0
                packet_model.total += engines['packets']

                packet_model.parse_failures = packet_model.parse_failures or 0
                packet_model.parse_failures += engines['parse_failures']

                packet_model.auth_failures = packet_model.auth_failures or 0
                packet_model.auth_failures += engines['auth_failures']

                packet_model.context_failures = packet_model.context_failures or 0
                packet_model.context_failures += engines['context_failures']

                for engine_id, security_models in engines.items():
                    if not isinstance(security_models, dict):
                        continue

                    for security_model, security_levels in security_models.items():
                        if not isinstance(security_levels, dict):
                            continue

                        for security_level, context_engines in security_levels.items():
                            if not isinstance(context_engines, dict):
                                continue

                            for context_engine_id, context_names in context_engines.items():
                                if not isinstance(context_names, dict):
                                    continue

                                for context_name, pdus in context_names.items():
                                    if not isinstance(pdus, dict):
                                        continue

                                    for pdu_type, recordings in pdus.items():
                                        if not isinstance(recordings, dict):
                                            continue

                                        for recording, counters in recordings.items():
                                            agent_model = models.Agent(
                                                transport_id=transport_model.id,
                                                engine=engine_id,
                                                security_model=security_model,
                                                security_level=security_level,
                                                context_engine=context_engine_id,
                                                context_name=context_name,
                                            )

                                            agent_model = db.session.merge(agent_model)

                                            autoincrement(agent_model, models.Agent)

                                            recording_model = models.Recording(
                                                agent_id=agent_model.id, path=recording)

                                            recording_model = db.session.merge(recording_model)

                                            autoincrement(recording_model, models.Recording)

                                            pdu_model = models.Pdu(
                                                recording_id=recording_model.id, name=pdu_type)

                                            pdu_model = db.session.merge(pdu_model)

                                            pdu_model.total = pdu_model.total or 0
                                            pdu_model.total += counters['pdus']

                                            autoincrement(pdu_model, models.Pdu)

                                            varbind_model = models.VarBind(pdu_id=pdu_model.id)

                                            varbind_model = db.session.merge(varbind_model)

                                            varbind_model.total = varbind_model.total or 0
                                            varbind_model.total += counters['varbinds']

                                            varbind_model.failures = varbind_model.failures or 0
                                            varbind_model.failures += counters['failures']

                                            variations = counters.get('variations', ())

                                            for name, counters in variations.items():

                                                variation_model = models.Variation(
                                                    pdu_id=pdu_model.id, name=name)

                                                variation_model = db.session.merge(variation_model)

                                                variation_model.total = variation_model.total or 0
                                                variation_model.total += counters['calls']

                                                variation_model.failures = variation_model.failures or 0
                                                variation_model.failures += counters['failures']

                                            db.session.commit()



#
# This file is part of SNMP simulator Control Plane software.
#
# Copyright (c) 2019-2020, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/snmpsim/license.html
#
# SNMP simulator metrics: REST API view schemas
#
import marshmallow
import flask

from snmpsim_control_plane.metrics import ma
from snmpsim_control_plane.metrics import models


class EnsureZeroMixIn(object):

    @marshmallow.pre_dump
    def ensure_zeros(self, item):
        return {key: getattr(item, key) or 0 for key in item.keys()}


class PacketsSchema(marshmallow.Schema, EnsureZeroMixIn):
    class Meta:
        fields = (
            'total', 'parse_failures', 'auth_failures',
            'context_failures')


class MessagesSchema(marshmallow.Schema, EnsureZeroMixIn):
    class Meta:
        fields = ('pdus', 'var_binds', 'failures', 'variations')

    variations = marshmallow.fields.Dict()


class VariationsSchema(marshmallow.Schema, EnsureZeroMixIn):
    class Meta:
        fields = ('name', 'total', 'failures')


class ConsoleSchema(ma.ModelSchema):
    class Meta:
        model = models.ConsolePage
        fields = ('id', 'timestamp', 'text', 'process', '_links')

    class ProcessSchema(ma.ModelSchema):
        class Meta:
            model = models.Process
            fields = ('id', 'path', '_links')

        _links = ma.Hyperlinks({
            'self': ma.URLFor('show_processes', id='<id>'),
        })

    process = ma.Nested(ProcessSchema)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_console', id='<process_id>', page_id='<id>'),
        'collection': ma.URLFor('show_console', id='<process_id>')
    })


class EndpointSchema(ma.ModelSchema):
    class Meta:
        model = models.Endpoint
        fields = ('id', 'protocol', 'address', 'process', '_links')

    class ProcessSchema(ma.ModelSchema):
        class Meta:
            model = models.Process
            fields = ('id', 'path', '_links')

        _links = ma.Hyperlinks({
            'self': ma.URLFor('show_processes', id='<id>'),
        })

    process = ma.Nested(ProcessSchema)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_endpoints', id='<process_id>', endpoint_id='<id>'),
        'collection': ma.URLFor('show_endpoints', id='<process_id>')
    })


class ProcessSchema(ma.ModelSchema):
    class Meta:
        model = models.Process
        fields = ('id', 'path', 'runtime', 'memory', 'cpu', 'files',
                  'exits', 'changes', 'last_update', 'update_interval',
                  'endpoints', 'supervisor', 'console_pages', '_links')

    class EndpointsSchema(ma.ModelSchema):
        class Meta:
            fields = ('count', '_links')

        count = marshmallow.fields.Method('get_count')
        _links = marshmallow.fields.Method('get_links')

        def get_count(self, endpoints):
            return len(endpoints)

        def get_links(self, endpoints):
            if endpoints:
                return {
                    'self': flask.url_for(
                        'show_endpoints', id=endpoints[0].process_id)
                }

            else:
                return {}

    endpoints = ma.Nested(EndpointsSchema)

    class ConsoleSchema(ma.ModelSchema):
        class Meta:
            model = models.ConsolePage
            fields = ('count', 'last_update', '_links')

        count = marshmallow.fields.Method('get_count')
        last_update = marshmallow.fields.Method('get_last_update')
        _links = marshmallow.fields.Method('get_links')

        def get_count(self, console_pages):
            return len(console_pages)

        def get_last_update(self, console_pages):
            if console_pages:
                return console_pages[-1].timestamp.isoformat() + '+00:00'

            return 0

        def get_links(self, console_pages):
            if console_pages:
                return {
                    'self': flask.url_for(
                        'show_console', id=console_pages[0].process_id)
                }

            else:
                return {}

    console_pages = ma.Nested(ConsoleSchema)

    class SupervisorSchema(ma.ModelSchema):
        class Meta:
            model = models.Supervisor
            fields = ('id', 'hostname', 'watch_dir', '_links')

        _links = ma.Hyperlinks({
            'self': ma.URLFor('show_supervisors', id='<id>'),
            'collection': ma.URLFor('show_supervisors')
        })

    supervisor = ma.Nested(SupervisorSchema)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_processes', id='<id>'),
        'collection': ma.URLFor('show_processes')
    })


class SupervisorSchema(ma.ModelSchema):
    class Meta:
        model = models.Supervisor
        fields = ('id', 'hostname', 'watch_dir', 'started',
                  'processes', '_links')

    class ProcessSchema(ma.ModelSchema):
        class Meta:
            model = models.Process
            fields = ('id', 'path', '_links')

        _links = ma.Hyperlinks({
            'self': ma.URLFor('show_processes', id='<id>'),
        })

    processes = ma.Nested(ProcessSchema, many=True)

    _links = ma.Hyperlinks({
        'self': ma.URLFor('show_supervisors', id='<id>'),
        'collection': ma.URLFor('show_supervisors')
    })
